import shutil
import os


def rmtree(folder='AutoML', must_exist=False):
    if must_exist and not os.path.exists(folder):
        raise FileNotFoundError(f'Directory {folder} does not exist')
    if os.path.exists(folder):
        shutil.rmtree(folder)



import os
import re
import json
import warnings

import requests
from pathlib import Path


__all__ = ['PlatformSynchronizer']


class PlatformSynchronizer:

    def __init__(
            self,
            connection_string_name='AMPLO_PLATFORM_STRING',
            api_url='https://platform.amplo.ch/api',
            verbose=0,
    ):
        """
        Connector to Amplo`s platform for uploading trained models

        Parameters
        ----------
        connection_string_name : str
        api_url : str
        verbose : int
        """
        self.api_url = re.sub(r'/$', '', api_url)
        self.api_key = os.getenv(connection_string_name)
        self.verbose = int(verbose)

    def upload_model(self, issue_dir, team, machine, service, issue, version):
        """
        Upload trained model

        Parameters
        ----------
        issue_dir : str or Path
            Model issue directory
        team : str
            Team name
        machine : str
            Machine name
        service : str
            Service name
        issue : str
            Issue name
        version : int
            Version of production model to upload

        Notes
        -----
        Data structures
            - Locally: ``AutoML / Production / vX / ...``
            - Cloud: ``Team / Machine / models / Diagnostics / vX / ...``
        """

        # Check input args
        assert os.path.isdir(issue_dir), 'Model directory does not exist'
        assert all(isinstance(arg, str) for arg in (team, machine, service, issue)), 'Invalid argument'
        assert isinstance(version, int), 'Invalid argument'

        # Check file structure
        file_dir = Path(issue_dir) / 'Production' / f'v{version}'
        assert os.path.isdir(file_dir), f'Production files for version {version} do not exist'
        files = [file_dir / file for file in ('Settings.json', 'Model.joblib', 'Report.pdf')]
        for file in files:
            assert file.exists(), f'File {file.name} does not exist'

        # Set info
        headers = {'X-Api-Key': self.api_key}
        data = {
            'team': team,  # team name
            'machine': machine,  # machine name
            'category': service,  # category
            'model': issue, 'name': issue,  # model name (for PUT or POST, respectively)
            'version': f'v{version}',  # latest version
        }

        # Assert that model exists
        def model_exists(model_name: str):
            # GET request
            get_response = requests.get(self.api_url + '/models', params=data, headers=headers)
            assert get_response.status_code == 200, 'GET request has failed'

            # Get all existing models
            try:
                # Convert string to list of dict
                existing_models = json.loads(get_response.text)
                if isinstance(existing_models, str):
                    ValueError('Could not transform response.')
            except json.decoder.JSONDecodeError as err:
                warnings.warn(f'Could not decode GET response - {err}')
                # Set default value of ``existing_models`` to list of (empty) dict
                existing_models = list(dict())
            assert isinstance(existing_models, list), 'Received unexpected data type'
            assert isinstance(existing_models[0], dict), 'Received unexpected data type'

            # Check if model exists
            return any(model.get('name', None) == model_name for model in existing_models)

        # POST when model yet doesn't exist
        if not model_exists(data['name']):
            warnings.warn(f'Model {data["name"]} did not yet exist. POST-ing a new model to the platform...')
            post_response = requests.post(self.api_url + '/models', data=data, headers=headers)
            # Assert that now model exists
            assert model_exists(data['name']), f'Failed to create model - {post_response} {post_response.text}'

        # Add files to data
        data.update({'files': [open(file, 'r') for file in files]})
        # PUT a new version
        put_response = requests.put(self.api_url + '/models', data=data, headers=headers)
        assert put_response.status_code == 200

    def upload_latest_model(self, issue_dir, *args, **kwargs):
        """
        Same as `upload_model` but selects the latest version of production model

        Parameters
        ----------
        issue_dir : str or Path
            Model issue directory
        args
            Arguments to be passed to `upload_model`
        kwargs
            Keyword arguments to be passed to `upload model`
        """
        # Check input args
        assert os.path.isdir(issue_dir), 'Model directory does not exist'

        # Check file structure
        all_version_dirs = [dir_ for dir_ in (Path(issue_dir) / 'Production').glob('v*') if dir_.is_dir()]
        assert len(all_version_dirs) > 0, 'No production data found'

        # Take the latest version and remove `v` in e.g. `v1`
        latest_version = int(sorted(all_version_dirs)[-1].name[1:])

        # Set version in kwargs and upload model
        kwargs.update({'version': latest_version})
        return self.upload_model(issue_dir, *args, **kwargs)



import pytest

from tests import rmtree


@pytest.fixture(autouse=True)
def rmtree_automl():
    folder = 'AutoML'
    rmtree(folder, must_exist=False)
    yield folder
    rmtree(folder, must_exist=False)



import re
import os
import time
import copy
import json
import Amplo
import joblib
import shutil
import warnings
import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Union
from pathlib import Path
from datetime import datetime
from shap import TreeExplainer
from shap import KernelExplainer

from sklearn import metrics
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold

from Amplo import Utils
from Amplo.AutoML.Sequencer import Sequencer
from Amplo.AutoML.Modeller import Modeller
from Amplo.AutoML.DataSampler import DataSampler
from Amplo.AutoML.DataExplorer import DataExplorer
from Amplo.AutoML.DataProcessor import DataProcessor
from Amplo.AutoML.DriftDetector import DriftDetector
from Amplo.AutoML.FeatureProcessor import FeatureProcessor
from Amplo.AutoML.IntervalAnalyser import IntervalAnalyser
from Amplo.Regressors.StackingRegressor import StackingRegressor
from Amplo.Classifiers.StackingClassifier import StackingClassifier

from .GridSearch.BaseGridSearch import BaseGridSearch
from .GridSearch.HalvingGridSearch import HalvingGridSearch
from .GridSearch.OptunaGridSearch import OptunaGridSearch

from .Documenting.MultiDocumenting import MultiDocumenting
from .Documenting.BinaryDocumenting import BinaryDocumenting
from .Documenting.RegressionDocumenting import RegressionDocumenting


class Pipeline:

    def __init__(self, **kwargs):
        """
        Automated Machine Learning Pipeline for tabular data.
        Designed for predictive maintenance applications, failure identification, failure prediction, condition
        monitoring, etc.

        Parameters
        ----------
        Main Parameters:
        main_dir [str]: Main directory of Pipeline (for documentation)
        target [str]: Column name of the output/dependent/regressand variable.
        name [str]: Name of the project (for documentation)
        version [int]: Pipeline version (set automatically)
        mode [str]: 'classification' or 'regression'
        objective [str]: from sklearn metrics and scoring

        Data Processor:
        int_cols [list[str]]: Column names of integer columns
        float_cols [list[str]]: Column names of float columns
        date_cols [list[str]]: Column names of datetime columns
        cat_cols [list[str]]: Column names of categorical columns
        missing_values [str]: [DataProcessing] - 'remove', 'interpolate', 'mean' or 'zero'
        outlier_removal [str]: [DataProcessing] - 'clip', 'boxplot', 'z-score' or 'none'
        z_score_threshold [int]: [DataProcessing] If outlier_removal = 'z-score', the threshold is adaptable
        include_output [bool]: Whether to include output in the training data (sensible only with sequencing)

        Feature Processor:
        extract_features [bool]: Whether to use FeatureProcessing module
        information_threshold : [FeatureProcessing] Threshold for removing co-linear features
        feature_timeout [int]: [FeatureProcessing] Time budget for feature processing
        max_lags [int]: [FeatureProcessing] Maximum lags for lagged features to analyse
        max_diff [int]: [FeatureProcessing] Maximum differencing order for differencing features

        Interval Analyser:
        interval_analyse [bool]: Whether to use IntervalAnalyser module
            Note that this has no effect when data from ``self._read_data`` is not multi-indexed

        Sequencing:
        sequence [bool]: [Sequencing] Whether to use Sequence module
        seq_back [int or list[int]]: Input time indices
            If list -> includes all integers within the list
            If int -> includes that many samples back
        seq_forward [int or list[int]: Output time indices
            If list -> includes all integers within the list.
            If int -> includes that many samples forward.
        seq_shift [int]: Shift input / output samples in time
        seq_diff [int]:  Difference the input & output, 'none', 'diff' or 'log_diff'
        seq_flat [bool]: Whether to return a matrix (True) or Tensor (Flat)

        Modelling:
        standardize [bool]: Whether to standardize input/output data
        shuffle [bool]: Whether to shuffle the samples during cross-validation
        cv_splits [int]: How many cross-validation splits to make
        store_models [bool]: Whether to store all trained model files

        Grid Search:
        grid_search_type [Optional[str]]: Which method to use 'optuna', 'halving', 'base' or None
        grid_search_time_budget : Time budget for grid search
        grid_search_candidates : Parameter evaluation budget for grid search
        grid_search_iterations : Model evaluation budget for grid search

        Stacking:
        stacking [bool]: Whether to create a stacking model at the end

        Production:
        preprocess_function [str]: Add custom code for the prediction function, useful for production. Will be executed
            with exec, can be multiline. Uses data as input.

        Flags:
        logging_level [Optional[Union[int, str]]]: Logging level for warnings, info, etc.
        plot_eda [bool]: Whether to run Exploratory Data Analysis
        process_data [bool]: Whether to force data processing
        document_results [bool]: Whether to force documenting
        no_dirs [bool]: Whether to create files or not
        verbose [int]: Level of verbosity
        """

        # Copy arguments
        ##################
        # Main Settings
        self.mainDir = kwargs.get('main_dir', 'AutoML/')
        self.target = re.sub('[^a-z0-9]', '_', kwargs.get('target', '').lower())
        self.name = kwargs.get('name', 'AutoML')
        self.version = kwargs.get('version', None)
        self.mode = kwargs.get('mode', None)
        self.objective = kwargs.get('objective', None)

        # Data Processor
        self.intCols = kwargs.get('int_cols', None)
        self.floatCols = kwargs.get('float_cols', None)
        self.dateCols = kwargs.get('date_cols', None)
        self.catCols = kwargs.get('cat_cols', None)
        self.missingValues = kwargs.get('missing_values', 'zero')
        self.outlierRemoval = kwargs.get('outlier_removal', 'clip')
        self.zScoreThreshold = kwargs.get('z_score_threshold', 4)
        self.includeOutput = kwargs.get('include_output', False)

        # Balancer
        self.balance = kwargs.get('balance', True)

        # Feature Processor
        self.extractFeatures = kwargs.get('extract_features', True)
        self.informationThreshold = kwargs.get('information_threshold', 0.999)
        self.featureTimeout = kwargs.get('feature_timeout', 3600)
        self.maxLags = kwargs.get('max_lags', 0)
        self.maxDiff = kwargs.get('max_diff', 0)

        # Interval Analyser
        self.useIntervalAnalyser = kwargs.get('interval_analyse', True)

        # Sequencer
        self.sequence = kwargs.get('sequence', False)
        self.sequenceBack = kwargs.get('seq_back', 1)
        self.sequenceForward = kwargs.get('seq_forward', 1)
        self.sequenceShift = kwargs.get('seq_shift', 0)
        self.sequenceDiff = kwargs.get('seq_diff', 'none')
        self.sequenceFlat = kwargs.get('seq_flat', True)

        # Modelling
        self.standardize = kwargs.get('standardize', False)
        self.shuffle = kwargs.get('shuffle', True)
        self.cvSplits = kwargs.get('cv_shuffle', 10)
        self.storeModels = kwargs.get('store_models', False)

        # Grid Search Parameters
        self.gridSearchType = kwargs.get('grid_search_type', 'optuna')
        self.gridSearchTimeout = kwargs.get('grid_search_time_budget', 3600)
        self.gridSearchCandidates = kwargs.get('grid_search_candidates', 250)
        self.gridSearchIterations = kwargs.get('grid_search_iterations', 3)

        # Stacking
        self.stacking = kwargs.get('stacking', False)

        # Production
        self.preprocessFunction = kwargs.get('preprocess_function', None)

        # Flags
        self.plotEDA = kwargs.get('plot_eda', False)
        self.processData = kwargs.get('process_data', True)
        self.documentResults = kwargs.get('document_results', True)
        self.verbose = kwargs.get('verbose', 0)
        self.noDirs = kwargs.get('no_dirs', False)

        # Checks
        assert self.mode in [None, 'regression', 'classification'], 'Supported modes: regression, classification.'
        assert 0 < self.informationThreshold < 1, 'Information threshold needs to be within [0, 1]'
        assert self.maxLags < 50, 'Max_lags too big. Max 50.'
        assert self.maxDiff < 5, 'Max diff too big. Max 5.'
        assert self.gridSearchType is None \
            or self.gridSearchType.lower() in ['base', 'halving', 'optuna'], \
            'Grid Search Type must be Base, Halving, Optuna or None'

        # Advices
        if self.includeOutput and not self.sequence:
            warnings.warn('[AutoML] IMPORTANT: strongly advices to not include output without sequencing.')

        # Create dirs
        if not self.noDirs:
            self._create_dirs()
            self._load_version()

        # Store Pipeline Settings
        self.settings = {'pipeline': kwargs, 'validation': {}, 'feature_set': ''}

        # Objective & Scorer
        self.scorer = None
        if self.objective is not None:
            assert isinstance(self.objective, str), 'Objective needs to be a string'
            assert self.objective in metrics.SCORERS.keys(), 'Metric not supported, look at sklearn.metrics'
            self.scorer = metrics.SCORERS[self.objective]

        # Required sub-classes
        self.dataSampler = DataSampler()
        self.dataProcessor = DataProcessor()
        self.dataSequencer = Sequencer()
        self.featureProcessor = FeatureProcessor()
        self.intervalAnalyser = IntervalAnalyser()
        self.driftDetector = DriftDetector()

        # Instance initiating
        self.bestModel = None
        self._data = None
        self.featureSets = None
        self.results = None
        self.n_classes = None
        self.is_fitted = False

        # Monitoring
        logging_level = kwargs.get('logging_level', 'INFO')
        logging_dir = Path(self.mainDir) / 'app_logs.log' if not self.noDirs else None
        self.logger = Utils.logging.get_logger('AutoML', logging_dir, logging_level, capture_warnings=True)
        self._prediction_time = None
        self._main_predictors = None

    # User Pointing Functions
    def get_settings(self, version: int = None) -> dict:
        """
        Get settings to recreate fitted object.

        Parameters
        ----------
        version : int, optional
            Production version, defaults to current version
        """
        if version is None or version == self.version:
            assert self.is_fitted, "Pipeline not yet fitted."
            return self.settings
        else:
            settings_path = self.mainDir + f'Production/v{self.version}/Settings.json'
            assert Path(settings_path).exists(), 'Cannot load settings from nonexistent version'
            return json.load(open(settings_path, 'r'))

    def load_settings(self, settings: dict):
        """
        Restores a pipeline from settings.

        Parameters
        ----------
        settings [dict]: Pipeline settings
        """
        # Set parameters
        settings['pipeline']['no_dirs'] = True
        self.__init__(**settings['pipeline'])
        self.settings = settings
        self.dataProcessor.load_settings(settings['data_processing'])
        self.featureProcessor.load_settings(settings['feature_processing'])
        # TODO: load_settings for IntervalAnalyser (not yet implemented)
        if 'drift_detector' in settings:
            self.driftDetector = DriftDetector(
                num_cols=self.dataProcessor.float_cols + self.dataProcessor.int_cols,
                cat_cols=self.dataProcessor.cat_cols,
                date_cols=self.dataProcessor.date_cols
            ).load_weights(settings['drift_detector'])

    def load_model(self, model: object):
        """
        Restores a trained model
        """
        assert type(model).__name__ == self.settings['model']
        self.bestModel = model
        self.is_fitted = True

    def fit(self, *args, **kwargs):
        """
        Fit the full AutoML pipeline.
            1. Prepare data for training
            2. Train / optimize models
            3. Prepare Production Files
                Nicely organises all required scripts / files to make a prediction

        Parameters
        ----------
        args
            For data reading - Propagated to `self.data_preparation`
        kwargs
            For data reading (propagated to `self.data_preparation`) AND
            for production filing (propagated to `self.conclude_fitting`)
        """
        # Starting
        print('\n\n*** Starting Amplo AutoML - {} ***\n\n'.format(self.name))

        # Prepare data for training
        self.data_preparation(*args, **kwargs)

        # Train / optimize models
        self.model_training()

        # Conclude fitting
        self.conclude_fitting(**kwargs)

    def data_preparation(self, *args, **kwargs):
        """
        Prepare data for modelling
            1. Data Processing
                Cleans all the data. See @DataProcessing
            2. (optional) Exploratory Data Analysis
                Creates a ton of plots which are helpful to improve predictions manually
            3. Feature Processing
                Extracts & Selects. See @FeatureProcessing

        Parameters
        ----------
        args
            For data reading - Propagated to `self._read_data`
        kwargs
            For data reading - Propagated to `self._read_data`
        """
        # Reading data
        self._read_data(*args, **kwargs)

        # Detect mode (classification / regression)
        self._mode_detector()

        # Preprocess Data
        self._data_processing()

        # Run Exploratory Data Analysis
        self._eda()

        # Balance data
        self._data_sampling()

        # Sequence
        self._sequencing()

        # Extract and select features
        self._feature_processing()

        # Interval-analyze data
        self._interval_analysis()

        # Standardize
        # Standardizing assures equal scales, equal gradients and no clipping.
        # Therefore, it needs to be after sequencing & feature processing, as this alters scales
        self._standardizing()

    def model_training(self):
        """
        Train models
            1. Initial Modelling
                Runs various off the shelf models with default parameters for all feature sets
                If Sequencing is enabled, this is where it happens, as here, the feature set is generated.
            2. Grid Search
                Optimizes the hyperparameters of the best performing models
            3. (optional) Create Stacking model
            4. (optional) Create documentation
        """
        # Run initial models
        self._initial_modelling()

        # Optimize Hyper parameters
        self.grid_search()

        # Create stacking model
        self._create_stacking()

    def conclude_fitting(self, *, model=None, feature_set=None, params=None, **kwargs):
        """
        Prepare production files that are necessary to deploy a specific
        model / feature set combination

        Creates or modifies the following files
            - ``Model.joblib`` (production model)
            - ``Settings.json`` (model settings)
            - ``Report.pdf`` (training report)

        Parameters
        ----------
        model : str, optional
            Model file for which to prepare production files
        feature_set : str, optional
            Feature set for which to prepare production files
        params : dict, optional
            Model parameters for which to prepare production files.
            Default: takes best parameters
        """
        # Set up production path
        prod_dir = self.mainDir + f'Production/v{self.version}/'
        Path(prod_dir).mkdir(exist_ok=True)

        # Parse arguments
        model, feature_set, params = self._parse_production_args(model, feature_set, params)

        # Verbose printing
        if self.verbose > 0:
            print(f'[AutoML] Preparing Production files for {model}, {feature_set}, {params}')

        # Set best model (`self.bestModel`)
        self._prepare_production_model(prod_dir + 'Model.joblib', model, feature_set, params)

        # Set and store production settings
        self._prepare_production_settings(prod_dir + 'Settings.json', model, feature_set, params)

        # Report
        report_path = self.mainDir + f'Documentation/v{self.version}/{model}_{feature_set}.pdf'
        if not Path(report_path).exists():
            self.document(self.bestModel, feature_set)
        shutil.copy(report_path, prod_dir + 'Report.pdf')

        # Finish
        self.is_fitted = True
        print('[AutoML] All done :)')

    def convert_data(self, x: pd.DataFrame, preprocess: bool = True) -> [pd.DataFrame, pd.Series]:
        """
        Function that uses the same process as the pipeline to clean data.
        Useful if pipeline is pickled for production

        Parameters
        ----------
        data [pd.DataFrame]: Input features
        """
        # Convert to Pandas
        if isinstance(x, np.ndarray):
            x = pd.DataFrame(x, columns=[f"Feature_{i}" for i in range(x.shape[1])])

        # Custom code
        if self.preprocessFunction is not None and preprocess:
            ex_globals = {'data': x}
            exec(self.preprocessFunction, ex_globals)
            x = ex_globals['data']

        # Process data
        x = self.dataProcessor.transform(x)

        # Drift Check
        self.driftDetector.check(x)

        # Split output
        y = None
        if self.target in x.keys():
            y = x[self.target]
            if not self.includeOutput:
                x = x.drop(self.target, axis=1)

        # Sequence
        if self.sequence:
            x, y = self.dataSequencer.convert(x, y)

        # Convert Features
        x = self.featureProcessor.transform(x, self.settings['feature_set'])

        # Standardize
        if self.standardize:
            x, y = self._transform_standardize(x, y)

        # NaN test -- datetime should be taken care of by now
        if x.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
            raise ValueError(f"Column(s) with NaN: {list(x.keys()[x.isna().sum() > 0])}")

        # Return
        return x, y

    def predict(self, data: pd.DataFrame) -> np.ndarray:
        """
        Full script to make predictions. Uses 'Production' folder with defined or latest version.

        Parameters
        ----------
        data [pd.DataFrame]: data to do prediction on
        """
        start_time = time.time()
        assert self.is_fitted, "Pipeline not yet fitted."

        # Print
        if self.verbose > 0:
            print('[AutoML] Predicting with {}, v{}'.format(type(self.bestModel).__name__, self.version))

        # Convert
        x, y = self.convert_data(data)

        # Predict
        if self.mode == 'regression' and self.standardize:
            predictions = self._inverse_standardize(self.bestModel.predict(x))
        else:
            predictions = self.bestModel.predict(x)

        # Stop timer
        self._prediction_time = (time.time() - start_time) / len(x) * 1000

        # Calculate main predictors
        self._get_main_predictors(x)

        return predictions

    def predict_proba(self, data: pd.DataFrame) -> np.ndarray:
        """
        Returns probabilistic prediction, only for classification.

        Parameters
        ----------
        data [pd.DataFrame]: data to do prediction on
        """
        start_time = time.time()
        assert self.is_fitted, "Pipeline not yet fitted."
        assert self.mode == 'classification', 'Predict_proba only available for classification'
        assert hasattr(self.bestModel, 'predict_proba'), '{} has no attribute predict_proba'.format(
            type(self.bestModel).__name__)

        # Print
        if self.verbose > 0:
            print('[AutoML] Predicting with {}, v{}'.format(type(self.bestModel).__name__, self.version))

        # Convert data
        x, y = self.convert_data(data)

        # Predict
        prediction = self.bestModel.predict_proba(x)

        # Stop timer
        self._prediction_time = (time.time() - start_time) / len(x) * 1000

        # Calculate main predictors
        self._get_main_predictors(x)

        return prediction

    # Fit functions
    def _read_data(self, x=None, y=None, *, data=None, **kwargs):
        """
        Reads and loads data into desired format.

        Expects to receive:
            1. Both, ``x`` and ``y`` (-> features and target), or
            2. Either ``x`` or ``data`` (-> dataframe or path to folder)

        Parameters
        ----------
        x : np.ndarray or pd.Series or pd.DataFrame or str or Path, optional
            x-data (input) OR acts as ``data`` parameter when param ``y`` is empty
        y : np.ndarray or pd.Series, optional
            y-data (target)
        data : pd.DataFrame or str or Path, optional
            Contains both, x and y, OR provides a path to folder structure

        Returns
        -------
        Pipeline
        """
        assert x is not None or data is not None, 'No data provided'
        assert (x is not None) ^ (data is not None), 'Setting both, `x` and `data`, is ambiguous'

        # Labels are provided separately
        if y is not None:
            # Check data
            x = x if x is not None else data
            assert x is not None, 'Parameter ``x`` is not set'
            assert isinstance(x, (np.ndarray, pd.Series, pd.DataFrame)), 'Unsupported data type for parameter ``x``'
            assert isinstance(y, (np.ndarray, pd.Series)), 'Unsupported data type for parameter ``y``'

            # Set target manually if not defined
            if self.target == '':
                self.target = 'target'

            # Parse x-data
            if isinstance(x, np.ndarray):
                x = pd.DataFrame(x)
            elif isinstance(x, pd.Series):
                x = pd.DataFrame(x)
            # Parse y-data
            if isinstance(y, np.ndarray):
                y = pd.Series(y, index=x.index)
            y.name = self.target

            # Check data
            assert all(x.index == y.index), '``x`` and ``y`` indices do not match'
            if self.target in x.columns:
                assert all(x[self.target] == y), 'Target column co-exists in both, ``x`` and ``y`` data, ' \
                                                 f'but has not equal content. Rename the column ``{self.target}`` ' \
                                                 'in ``x`` or set a (different) target in initialization.'

            # Concatenate x and y
            data = pd.concat([x, y], axis=1)

        # Set data parameter in case it is provided through parameter ``x``
        data = data if data is not None else x
        metadata = None

        # A path was provided to read out (multi-indexed) data
        if isinstance(data, (str, Path)):
            # Set target manually if not defined
            if self.target == '':
                self.target = 'target'
            # Parse data
            data, metadata = Utils.io.merge_logs(data, self.target)

        # Test data
        assert self.target != '', 'No target string provided'
        assert self.target in data.columns, 'Target column missing'
        assert len(data.columns) == data.columns.nunique(), 'Columns have no unique names'

        # Parse data
        y = data[self.target]
        x = data.drop(self.target, axis=1)
        if isinstance(x.columns, pd.RangeIndex):
            x.columns = [f'Feature_{i}' for i in range(x.shape[1])]
        # Concatenate x and y
        data = pd.concat([x, y], axis=1)

        # Save data
        self.set_data(data)

        # Store metadata in settings
        file_metadata = dict()
        if metadata is not None:
            for _, row in metadata.iterrows():
                folder = row.pop('folder')
                file = row.pop('file')
                # Set metadata
                file_metadata[f'{folder}/{file}'] = row.to_dict()
        self.settings['file_metadata'] = file_metadata

        return self

    def has_new_training_data(self):
        # Return True if no previous version exists
        if self.version == 1:
            return True

        # Get previous and current file metadata
        curr_metadata = self.settings['file_metadata']
        last_metadata = self.get_settings(self.version - 1)['files']
        # Return True if metadata do not match
        if set(curr_metadata.keys()) != set(last_metadata.keys()):
            return True
        # Check modification date of each file
        return any(curr_metadata[file]['last_modified'] != last_metadata[file]['last_modified']
                   for file in curr_metadata.keys())

    def _mode_detector(self):
        """
        Detects the mode (Regression / Classification)
        """
        # Only run if mode is not provided
        if self.mode is None:

            # Classification if string
            if self.y.dtype == str or self.y.nunique() < 0.1 * len(self.data):
                self.mode = 'classification'
                self.objective = 'neg_log_loss'

            # Else regression
            else:
                self.mode = 'regression'
                self.objective = 'neg_mean_absolute_error'
            self.scorer = metrics.SCORERS[self.objective]

            # Copy to settings
            self.settings['pipeline']['mode'] = self.mode
            self.settings['pipeline']['objective'] = self.objective

            # Print
            if self.verbose > 0:
                print(f"[AutoML] Setting mode to {self.mode} & objective to {self.objective}.")
        return

    def _data_processing(self):
        """
        Organises the data cleaning. Heavy lifting is done in self.dataProcessor, but settings etc. needs
        to be organised.
        """
        self.dataProcessor = DataProcessor(target=self.target, int_cols=self.intCols, float_cols=self.floatCols,
                                           date_cols=self.dateCols, cat_cols=self.catCols,
                                           missing_values=self.missingValues,
                                           outlier_removal=self.outlierRemoval, z_score_threshold=self.zScoreThreshold)

        # Set paths
        data_path = self.mainDir + f'Data/Cleaned_v{self.version}.csv'
        settings_path = self.mainDir + f'Settings/Cleaning_v{self.version}.json'

        if Path(data_path).exists() and Path(settings_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self.set_data(data)

            # Load settings
            self.settings['data_processing'] = json.load(open(settings_path, 'r'))
            self.dataProcessor.load_settings(self.settings['data_processing'])

            if self.verbose > 0:
                print('[AutoML] Loaded Cleaned Data')

        else:
            # Cleaning
            data = self.dataProcessor.fit_transform(self.data)
            self.set_data(data)

            # Store data
            self._write_csv(self.data, data_path)

            # Save settings
            self.settings['data_processing'] = self.dataProcessor.get_settings()
            json.dump(self.settings['data_processing'], open(settings_path, 'w'))

        # If no columns were provided, load them from data processor
        if self.dateCols is None:
            self.dateCols = self.settings['data_processing']['date_cols']
        if self.intCols is None:
            self.dateCols = self.settings['data_processing']['int_cols']
        if self.floatCols is None:
            self.floatCols = self.settings['data_processing']['float_cols']
        if self.catCols is None:
            self.catCols = self.settings['data_processing']['cat_cols']

        # Assert classes in case of classification
        self.n_classes = self.y.nunique()
        if self.mode == 'classification':
            if self.n_classes >= 50:
                warnings.warn('More than 20 classes, you may want to reconsider classification mode')
            if set(self.y) != set([i for i in range(len(set(self.y)))]):
                raise ValueError('Classes should be [0, 1, ...]')

    def _eda(self):
        if self.plotEDA:
            print('[AutoML] Starting Exploratory Data Analysis')
            eda = DataExplorer(self.x, y=self.y,
                               mode=self.mode,
                               folder=self.mainDir,
                               version=self.version)
            eda.run()

    def _data_sampling(self):
        """
        Only run for classification problems. Balances the data using imblearn.
        Does not guarantee to return balanced classes. (Methods are data dependent)
        """
        self.dataSampler = DataSampler(method='both', margin=0.1, cv_splits=self.cvSplits, shuffle=self.shuffle,
                                       fast_run=False, objective=self.objective)

        # Set paths
        data_path = self.mainDir + f'Data/Balanced_v{self.version}.csv'

        # Only necessary for classification
        if self.mode == 'classification' and self.balance:

            if Path(data_path).exists():
                # Load data
                data = self._read_csv(data_path)
                self.set_data(data)

                if self.verbose > 0:
                    print('[AutoML] Loaded Balanced data')

            else:
                # Fit and resample
                print('[AutoML] Resampling data')
                x, y = self.dataSampler.fit_resample(self.x, self.y)

                # Store
                self._set_xy(x, y)
                self._write_csv(self.data, data_path)

    def _sequencing(self):
        """
        Sequences the data. Useful mostly for problems where older samples play a role in future values.
        The settings of this module are NOT AUTOMATIC
        """
        self.dataSequencer = Sequencer(back=self.sequenceBack, forward=self.sequenceForward,
                                       shift=self.sequenceShift, diff=self.sequenceDiff)

        # Set paths
        data_path = self.mainDir + f'Data/Sequence_v{self.version}.csv'

        if self.sequence:

            if Path(data_path).exists():
                # Load data
                data = self._read_csv(data_path)
                self.set_data(data)

                if self.verbose > 0:
                    print('[AutoML] Loaded Extracted Features')

            else:
                # Sequencing
                print('[AutoML] Sequencing data')
                x, y = self.dataSequencer.convert(self.x, self.y)

                # Store
                self._set_xy(x, y)
                self._write_csv(self.data, data_path)

    def _feature_processing(self):
        """
        Organises feature processing. Heavy lifting is done in self.featureProcessor, but settings, etc.
        needs to be organised.
        """
        self.featureProcessor = FeatureProcessor(mode=self.mode, max_lags=self.maxLags, max_diff=self.maxDiff,
                                                 extract_features=self.extractFeatures, timeout=self.featureTimeout,
                                                 information_threshold=self.informationThreshold)

        # Set paths
        data_path = self.mainDir + f'Data/Extracted_v{self.version}.csv'
        settings_path = self.mainDir + f'Settings/Extracting_v{self.version}.json'

        if Path(data_path).exists() and Path(settings_path).exists():
            # Loading data
            x = self._read_csv(data_path)
            self._set_x(x)

            # Loading settings
            self.settings['feature_processing'] = json.load(open(settings_path, 'r'))
            self.featureProcessor.load_settings(self.settings['feature_processing'])
            self.featureSets = self.settings['feature_processing']['featureSets']

            if self.verbose > 0:
                print('[AutoML] Loaded Extracted Features')

        else:
            print('[AutoML] Starting Feature Processor')

            # Transform data
            x, self.featureSets = self.featureProcessor.fit_transform(self.x, self.y)

            # Store data
            self._set_x(x)
            self._write_csv(self.x, data_path)

            # Save settings
            self.settings['feature_processing'] = self.featureProcessor.get_settings()
            json.dump(self.settings['feature_processing'], open(settings_path, 'w'))

    def _interval_analysis(self):
        """
        Interval-analyzes the data using ``Amplo.AutoML.IntervalAnalyser``
        or resorts to pre-computed data, if present.
        """
        # Skip analysis when analysis is not possible and/or not desired
        is_interval_analyzable = len(self.x.index.names) == 2
        if not (self.useIntervalAnalyser and is_interval_analyzable):
            return

        self.intervalAnalyser = IntervalAnalyser(target=self.target)

        # Set paths
        data_path = self.mainDir + f'Data/Interval_Analyzed_v{self.version}.csv'
        settings_path = self.mainDir + f'Settings/Interval_Analysis_v{self.version}.json'

        if Path(data_path).exists():  # TODO: and Path(settings_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self.set_data(data)

            # TODO implement `IntervalAnalyser.load_settings` and add to `self.load_settings`
            # # Load settings
            # self.settings['interval_analysis'] = json.load(open(settings_path, 'r'))
            # self.intervalAnalyser.load_settings(self.settings['interval_analysis'])

            if self.verbose > 0:
                print('[AutoML] Loaded interval-analyzed data')

        else:
            print(f'[AutoML] Interval-analyzing data')

            # Transform data
            data = self.intervalAnalyser.fit_transform(self.x, self.y)

            # Store data
            self.set_data(data)
            self._write_csv(self.data, data_path)

            # TODO implement `IntervalAnalyser.get_settings` and add to `self.get_settings`
            # # Save settings
            # self.settings['interval_analysis'] = self.intervalAnalyser.get_settings()
            # json.dump(self.settings['interval_analysis'], open(settings_path, 'w'))

    def _standardizing(self):
        """
        Wrapper function to determine whether to fit or load
        """
        # Return if standardize is off
        if not self.standardize:
            return

        # Set paths
        settings_path = self.mainDir + f'Settings/Standardize_v{self.version}.json'

        if Path(settings_path).exists():
            # Load data
            self.settings['standardize'] = json.load(open(settings_path, 'r'))

        else:
            # Fit data
            self._fit_standardize(self.x, self.y)

            # Store Settings
            json.dump(self.settings['standardize'], open(settings_path, 'w'))

        # Transform data
        x, y = self._transform_standardize(self.x, self.y)
        self._set_xy(x, y)

    def _initial_modelling(self):
        """
        Runs various models to see which work well.
        """

        # Set paths
        results_path = Path(self.mainDir) / 'Results.csv'

        # Load existing results
        if results_path.exists():

            # Load results
            self.results = pd.read_csv(results_path)

            # Printing here as we load it
            results = self.results[np.logical_and(
                self.results['version'] == self.version,
                self.results['type'] == 'Initial modelling'
            )]
            for fs in set(results['dataset']):
                print(f'[AutoML] Initial Modelling for {fs} ({len(self.featureSets[fs])})')
                fsr = results[results['dataset'] == fs]
                for i in range(len(fsr)):
                    row = fsr.iloc[i]
                    print(f'[AutoML] {row["model"].ljust(40)} {self.objective}: '
                          f'{row["mean_objective"]:.4f} \u00B1 {row["std_objective"]:.4f}')

        # Check if this version has been modelled
        if self.results is None or self.version not in self.results['version'].values:

            # Iterate through feature sets
            for feature_set, cols in self.featureSets.items():

                # Skip empty sets
                if len(cols) == 0:
                    print(f'[AutoML] Skipping {feature_set} features, empty set')
                    continue
                print(f'[AutoML] Initial Modelling for {feature_set} features ({len(cols)})')

                # Do the modelling
                modeller = Modeller(mode=self.mode, shuffle=self.shuffle, store_models=self.storeModels,
                                    objective=self.objective, dataset=feature_set,
                                    store_results=False, folder=self.mainDir + 'Models/')
                results = modeller.fit(self.x[cols], self.y)

                # Add results to memory
                results['type'] = 'Initial modelling'
                results['version'] = self.version
                if self.results is None:
                    self.results = results
                else:
                    self.results = pd.concat([self.results, results])

            # Save results
            self.results.to_csv(results_path, index=False)

    def grid_search(self, model=None, feature_set: str = None, parameter_set: str = None):
        """
        Runs a grid search. By default, takes the self.results, and runs for the top 3 optimizations.
        There is the option to provide a model & feature_set, but both have to be provided. In this case,
        the model & data set combination will be optimized.
        Implemented types, Base, Halving, Optuna

        Parameters
        ----------
        model [Object or str]- (optional) Which model to run grid search for.
        feature_set [str]- (optional) Which feature set to run grid search for 'rft', 'rfi' or 'pps'
        parameter_set [dict]- (optional) Parameter grid to optimize over
        """

        assert (model is not None and feature_set is not None) or model == feature_set, \
            'Model & feature_set need to be either both None or both provided.'

        # Skip grid search and set best initial model as best grid search parameters
        if self.gridSearchType is None or self.gridSearchIterations == 0:
            best_initial_model = self._sort_results(self.results[self.results['version'] == self.version]).iloc[0]
            best_initial_model['type'] = 'Hyper Parameter'
            self.results = pd.concat([self.results, best_initial_model.to_frame().T], ignore_index=True)
            return

        # If arguments are provided
        if model is not None:

            # Get model string
            if isinstance(model, str):
                model = Utils.utils.get_model(model, mode=self.mode, samples=len(self.x))

            # Organise existing results
            results = self.results[np.logical_and(
                self.results['model'] == type(model).__name__,
                self.results['version'] == self.version,
            )]
            results = self._sort_results(results[results['dataset'] == feature_set])

            # Check if exists and load
            if ('Hyper Parameter' == results['type']).any():
                print('[AutoML] Loading optimization results.')
                hyper_opt_results = results[results['type'] == 'Hyper Parameter']
                params = Utils.io.parse_json(hyper_opt_results.iloc[0]['params'])

            # Or run
            else:
                # Run grid search
                grid_search_results = self._sort_results(self._grid_search_iteration(model, parameter_set, feature_set))

                # Store results
                grid_search_results['model'] = type(model).__name__
                grid_search_results['version'] = self.version
                grid_search_results['dataset'] = feature_set
                grid_search_results['type'] = 'Hyper Parameter'
                self.results = self.results.append(grid_search_results)
                self.results.to_csv(self.mainDir + 'Results.csv', index=False)

                # Get params for validation
                params = Utils.io.parse_json(grid_search_results.iloc[0]['params'])

            # Validate
            if self.documentResults:
                self.document(model.set_params(**params), feature_set)
            return

        # If arguments aren't provided, run through promising models
        results = self._sort_results(self.results[np.logical_and(
            self.results['type'] == 'Initial modelling',
            self.results['version'] == self.version,
        )])
        for iteration in range(self.gridSearchIterations):
            # Grab settings
            settings = results.iloc[iteration]  # IndexError
            model = Utils.utils.get_model(settings['model'], mode=self.mode, samples=len(self.x))
            feature_set = settings['dataset']

            # Check whether exists
            model_results = self.results[np.logical_and(
                self.results['model'] == type(model).__name__,
                self.results['version'] == self.version,
            )]
            model_results = self._sort_results(model_results[model_results['dataset'] == feature_set])

            # If exists
            if ('Hyper Parameter' == model_results['type']).any():
                hyper_opt_res = model_results[model_results['type'] == 'Hyper Parameter']
                params = Utils.io.parse_json(hyper_opt_res.iloc[0]['params'])

            # Else run
            else:
                # For one model
                grid_search_results = self._sort_results(self._grid_search_iteration(
                    copy.deepcopy(model), parameter_set, feature_set))

                # Store
                grid_search_results['version'] = self.version
                grid_search_results['dataset'] = feature_set
                grid_search_results['type'] = 'Hyper Parameter'
                self.results = self.results.append(grid_search_results)
                self.results.to_csv(self.mainDir + 'Results.csv', index=False)
                params = Utils.io.parse_json(grid_search_results.iloc[0]['params'])

            # Validate
            if self.documentResults:
                self.document(model.set_params(**params), feature_set)

    def _create_stacking(self):
        """
        Based on the best performing models, in addition to cheap models based on very different assumptions,
        A stacking model is optimized to enhance/combine the performance of the models.
        --> should contain a large variety of models
        --> classifiers need predict_proba
        --> level 1 needs to be ordinary least squares
        """
        if self.stacking:
            print('[AutoML] Creating Stacking Ensemble')

            # Select feature set that has been picked most often for hyper parameter optimization
            results = self._sort_results(self.results[np.logical_and(
                self.results['type'] == 'Hyper Parameter',
                self.results['version'] == self.version,
            )])
            feature_set = results['dataset'].value_counts().index[0]
            results = results[results['dataset'] == feature_set]
            print('[AutoML] Selected Stacking feature set: {}'.format(feature_set))

            # Create Stacking Model Params
            n_stacking_models = 3
            stacking_models_str = results['model'].unique()[:n_stacking_models]
            stacking_models_params = [Utils.io.parse_json(results.iloc[np.where(results['model'] == sms)[0][0]]['params'])
                                      for sms in stacking_models_str]
            stacking_models = dict([(sms, stacking_models_params[i]) for i, sms in enumerate(stacking_models_str)])
            print('[AutoML] Stacked models: {}'.format(list(stacking_models.keys())))

            # Add samples & Features
            stacking_models['n_samples'], stacking_models['n_features'] = self.x.shape

            # Prepare Stack
            if self.mode == 'regression':
                stack = StackingRegressor(**stacking_models)
                cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)

            elif self.mode == 'classification':
                stack = StackingClassifier(**stacking_models)
                cv = StratifiedKFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            else:
                raise NotImplementedError('Unknown mode')

            # Cross Validate
            x, y = self.x[self.featureSets[feature_set]].to_numpy(), self.y.to_numpy()
            score = []
            times = []
            for (t, v) in tqdm(cv.split(x, y)):
                start_time = time.time()
                xt, xv, yt, yv = x[t], x[v], y[t].reshape((-1)), y[v].reshape((-1))
                model = copy.deepcopy(stack)
                model.fit(xt, yt)
                score.append(self.scorer(model, xv, yv))
                times.append(time.time() - start_time)

            # Output Results
            print('[AutoML] Stacking result:')
            print('[AutoML] {}:        {:.2f} \u00B1 {:.2f}'.format(self.objective, np.mean(score), np.std(score)))
            self.results = self.results.append({
                'date': datetime.today().strftime('%d %b %y'),
                'model': type(stack).__name__,
                'dataset': feature_set,
                'params': json.dumps(stack.get_params()),
                'mean_objective': np.mean(score),
                'std_objective': np.std(score),
                'mean_time': np.mean(times),
                'std_time': np.std(times),
                'version': self.version,
                'type': 'Stacking',
            }, ignore_index=True)
            self.results.to_csv(self.mainDir + 'Results.csv', index=False)

            # Document
            if self.documentResults:
                self.document(stack, feature_set)

    def document(self, model, feature_set: str):
        """
        Loads the model and features and initiates the outside Documenting class.

        Parameters
        ----------
        model [Object or str]- (optional) Which model to run grid search for.
        feature_set [str]- (optional) Which feature set to run grid search for 'rft', 'rfi' or 'pps'
        """
        # Get model
        if isinstance(model, str):
            model = Utils.utils.get_model(model, mode=self.mode, samples=len(self.x))

        # Checks
        assert feature_set in self.featureSets.keys(), 'Feature Set not available.'
        if os.path.exists(self.mainDir + 'Documentation/v{}/{}_{}.pdf'.format(
                self.version, type(model).__name__, feature_set)):
            print('[AutoML] Documentation existing for {} v{} - {} '.format(
                type(model).__name__, self.version, feature_set))
            return
        if len(model.get_params()) == 0:
            warnings.warn('[Documenting] Supplied model has no parameters!')

        # Run validation
        print('[AutoML] Creating Documentation for {} - {}'.format(type(model).__name__, feature_set))
        if self.mode == 'classification' and self.n_classes == 2:
            documenting = BinaryDocumenting(self)
        elif self.mode == 'classification':
            documenting = MultiDocumenting(self)
        elif self.mode == 'regression':
            documenting = RegressionDocumenting(self)
        else:
            raise ValueError('Unknown mode.')
        documenting.create(model, feature_set)

        # Append to settings
        self.settings['validation']['{}_{}'.format(type(model).__name__, feature_set)] = documenting.outputMetrics

    def _parse_production_args(self, model=None, feature_set=None, params=None):
        if model is not None and not isinstance(model, str):
            # TODO: This issue is linked with AML-103 (in Jira)
            #  1. Add to method docstring that it accepts a model instance, too
            #  2. Change `if`-case to a single `isinstance(model, BasePredictor)`
            # Get model name
            model = type(model).__name__

        # Get results of current version
        results = self._sort_results(self.results[self.results['version'] == self.version])

        if model is not None:
            # Filter results
            results = self._sort_results(results[results['model'] == model])

        if feature_set is not None:
            # Filter results
            results = self._sort_results(results[results['dataset'] == feature_set])

        if params is None:
            # Get best parameters
            params = results.iloc[0]['params']

        # Find the best allowed arguments
        model = model or results.iloc[0]['model']
        feature_set = feature_set or results.iloc[0]['dataset']
        params = Utils.io.parse_json(params or results.iloc[0]['params'])

        return model, feature_set, params

    def _prepare_production_model(self, model_path, model, feature_set, params):
        """
        Prepare and store `self.bestModel` for production

        Parameters
        ----------
        model_path : str or Path
            Where to store model for production
        model : str, optional
            Model file for which to prepare production files
        feature_set : str, optional
            Feature set for which to prepare production files
        params : dict, optional
            Model parameters for which to prepare production files.
            Default: takes best parameters

        Returns
        -------
        (str, str)
            Updated model name and feature set name
        """

        model_path = Path(model_path)

        # Try to load model from file
        if model_path.exists():
            # Load model
            self.bestModel = joblib.load(Path(model_path))
            # Reset if it's not the desired model
            if type(self.bestModel).__name__ != model or self.bestModel.get_params() != params:
                self.bestModel = None
        else:
            self.bestModel = None

        # Set best model
        if self.bestModel is not None:
            if self.verbose > 0:
                print('[AutoML] Loading existing model file')

        else:
            # Make model
            if 'Stacking' in model:
                # Create stacking
                if self.mode == 'regression':
                    self.bestModel = StackingRegressor(n_samples=len(self.x), n_features=len(self.x.keys()))
                elif self.mode == 'classification':
                    self.bestModel = StackingClassifier(n_samples=len(self.x), n_features=len(self.x.keys()))
                else:
                    raise NotImplementedError("Mode not set")
            else:
                # Take model as is
                self.bestModel = Utils.utils.get_model(model, mode=self.mode, samples=len(self.x))

            # Set params, train
            self.bestModel.set_params(**params)
            self.bestModel.fit(self.x[self.featureSets[feature_set]], self.y)

            # Save model
            joblib.dump(self.bestModel, model_path)

            if self.verbose > 0:
                score = self.scorer(self.bestModel, self.x[self.featureSets[feature_set]], self.y)
                print(f'[AutoML] Model fully fitted, in-sample {self.objective}: {score:4f}')

        return model, feature_set

    def _prepare_production_settings(self, settings_path, model=None, feature_set=None, params=None):
        """
        Prepare `self.settings` for production and dump to file

        Parameters
        ----------
        settings_path : str or Path
            Where to save settings for production
        model : str, optional
            Model file for which to prepare production files
        feature_set : str, optional
            Feature set for which to prepare production files
        params : dict, optional
            Model parameters for which to prepare production files.
            Default: takes best parameters
        """
        assert self.bestModel is not None, '`self.bestModel` is not yet prepared'
        settings_path = Path(settings_path)

        # Update pipeline settings
        self.settings['version'] = self.version
        self.settings['pipeline']['verbose'] = self.verbose
        self.settings['model'] = model
        self.settings['params'] = params
        self.settings['feature_set'] = feature_set
        self.settings['features'] = self.featureSets[feature_set]
        self.settings['amplo_version'] = Amplo.__version__ if hasattr(Amplo, '__version__') else 'dev'

        # Prune Data Processor
        required_features = self.featureProcessor.get_required_features(self.featureSets[feature_set])
        self.dataProcessor.prune_features(required_features)
        self.settings['data_processing'] = self.dataProcessor.get_settings()

        # Fit Drift Detector
        self.driftDetector = DriftDetector(
            num_cols=self.dataProcessor.float_cols + self.dataProcessor.int_cols,
            cat_cols=self.dataProcessor.cat_cols,
            date_cols=self.dataProcessor.date_cols
        )
        self.driftDetector.fit(self.x)
        self.driftDetector.fit_output(self.bestModel, self.x[self.featureSets[feature_set]])
        self.settings['drift_detector'] = self.driftDetector.get_weights()

        # Save settings
        json.dump(self.settings, open(settings_path, 'w'), indent=4)

    # Getter Functions / Properties
    @property
    def data(self) -> Union[None, pd.DataFrame]:
        return self._data

    @property
    def x(self) -> pd.DataFrame:
        if self.data is None:
            raise AttributeError('Data is None')
        if self.includeOutput:
            return self.data
        return self.data.drop(self.target, axis=1)

    @property
    def y(self):
        if self.data is None:
            raise AssertionError('`self.data` is empty. Set a value with `set_data`')
        return self.data[self.target]

    @property
    def y_orig(self):
        enc_labels = self.y
        dec_labels = self.dataProcessor.decode_labels(enc_labels, except_not_fitted=False)
        return pd.Series(dec_labels, name=self.target, index=enc_labels.index)

    # Setter Functions
    def set_data(self, new_data: pd.DataFrame):
        assert isinstance(new_data, pd.DataFrame), 'Invalid data type'
        assert self.target in new_data, 'No target column present'
        assert len(new_data.columns) > 1, 'No feature column present'
        self._data = new_data

    def _set_xy(self, new_x: Union[np.ndarray, pd.DataFrame], new_y: Union[np.ndarray, pd.Series]):
        if not isinstance(new_y, pd.Series):
            new_y = pd.Series(new_y, name=self.target)
        new_data = pd.concat([new_x, new_y], axis=1)
        self.set_data(new_data)

    def _set_x(self, new_x: Union[np.ndarray, pd.DataFrame]):
        old_x = self.data.drop(self.target, axis=1)
        # Convert to dataframe
        if isinstance(new_x, np.ndarray) and new_x.shape == old_x.shape:
            new_x = pd.DataFrame(new_x, index=old_x.index, columns=old_x.columns)
        elif isinstance(new_x, np.ndarray):
            warnings.warn('Old x-data has more/less columns than new x-data. Setting dummy feature names...')
            new_x = pd.DataFrame(new_x, index=old_x.index, columns=[f'Feature_{i}' for i in range(new_x.shape[1])])
        # Assert that target is not in x-data
        if self.target in new_x:
            raise AttributeError('Target column name should not be in x-data')
        self._data = pd.concat([new_x, self.y], axis=1)

    def set_y(self, new_y: Union[np.ndarray, pd.Series]):
        self._data[self.target] = new_y

    # Support Functions
    @staticmethod
    def _read_csv(data_path) -> pd.DataFrame:
        """
        Read data from given path and set index or multi-index

        Parameters
        ----------
        data_path : str or Path
        """
        assert Path(data_path).suffix == '.csv', 'Expected a *.csv path'

        data = pd.read_csv(data_path)

        if {'index', 'log'}.issubset(data.columns):
            # Multi-index: case when IntervalAnalyser was used
            index = ['log', 'index']
        elif 'index' in data.columns:
            index = ['index']
        else:
            raise IndexError('No known index was found. '
                             'Expected to find at least a column named `index`.')
        return data.set_index(index)

    @staticmethod
    def _write_csv(data, data_path):
        """
        Write data to given path and set index if needed.

        Parameters
        ----------
        data : pd.DataFrame or pd.Series
        data_path : str or Path
        """
        assert Path(data_path).suffix == '.csv', 'Expected a *.csv path'

        # Set single-index if not already present
        if len(data.index.names) == 1 and data.index.name is None:
            data.index.name = 'index'
        # Raise error if unnamed index is present
        if None in data.index.names:
            raise IndexError(f'Found an unnamed index column ({list(data.index.names)}).')

        # Write data
        data.to_csv(data_path)

    def _load_version(self):
        """
        Upon start, loads version
        """
        # No need if version is set
        if self.version is not None:
            return

        # Read changelog (if existent)
        if os.path.exists(self.mainDir + 'changelog.txt'):
            with open(self.mainDir + 'changelog.txt', 'r') as f:
                changelog = f.read()
        else:
            changelog = ''

        # Find production folders
        completed_versions = len(os.listdir(self.mainDir + 'Production'))
        started_versions = len(changelog.split('\n')) - 1

        # For new runs
        if started_versions == 0:
            with open(self.mainDir + 'changelog.txt', 'w') as f:
                f.write('v1: Initial Run\n')
            self.version = 1

        # If last run was completed and we start a new
        elif started_versions == completed_versions and self.processData:
            self.version = started_versions + 1
            with open(self.mainDir + 'changelog.txt', 'a') as f:
                notice = input(f'Changelog v{self.version}:\n')
                f.write(f'v{self.version}: {notice}\n')

        # If no new run is started (either continue or rerun)
        else:
            self.version = started_versions

        if self.verbose > 0:
            print(f'[AutoML] Setting Version {self.version}')

    def _create_dirs(self):
        folders = ['', 'EDA', 'Data', 'Features', 'Documentation', 'Production', 'Settings']
        for folder in folders:
            if not os.path.exists(self.mainDir + folder):
                os.makedirs(self.mainDir + folder)

    def sort_results(self, results: pd.DataFrame) -> pd.DataFrame:
        return self._sort_results(results)

    def _fit_standardize(self, x: pd.DataFrame, y: pd.Series):
        """
        Fits a standardization parameters and returns the transformed data
        """

        # Fit Input
        cat_cols = [k for lst in self.settings['data_processing']['dummies'].values() for k in lst]
        features = [k for k in x.keys() if k not in self.dateCols and k not in cat_cols]
        means_ = x[features].mean(axis=0)
        stds_ = x[features].std(axis=0)
        stds_[stds_ == 0] = 1
        settings = {
            'input': {
                'features': features,
                'means': means_.to_list(),
                'stds': stds_.to_list(),
            }
        }

        # Fit Output
        if self.mode == 'regression':
            std = y.std()
            settings['output'] = {
                'mean': y.mean(),
                'std': std if std != 0 else 1,
            }

        self.settings['standardize'] = settings

    def _transform_standardize(self, x: pd.DataFrame, y: pd.Series) -> [pd.DataFrame, pd.Series]:
        """
        Standardizes the input and output with values from settings.

        Parameters
        ----------
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data
        """
        # Input
        assert self.settings['standardize'], "Standardize settings not found."

        # Pull from settings
        features = self.settings['standardize']['input']['features']
        means = self.settings['standardize']['input']['means']
        stds = self.settings['standardize']['input']['stds']

        # Filter if not all features are present
        if len(x.keys()) < len(features):
            indices = [[i for i, j in enumerate(features) if j == k][0] for k in x.keys()]
            features = [features[i] for i in indices]
            means = [means[i] for i in indices]
            stds = [stds[i] for i in indices]

        # Transform Input
        x[features] = (x[features] - means) / stds

        # Transform output (only with regression)
        if self.mode == 'regression':
            y = (y - self.settings['standardize']['output']['mean']) / self.settings['standardize']['output']['std']

        return x, y

    def _inverse_standardize(self, y: pd.Series) -> pd.Series:
        """
        For predictions, transform them back to application scales.
        Parameters
        ----------
        y [pd.Series]: Standardized output

        Returns
        -------
        y [pd.Series]: Actual output
        """
        assert self.settings['standardize'], "Standardize settings not found"
        return y * self.settings['standardize']['output']['std'] + self.settings['standardize']['output']['mean']

    @staticmethod
    def _sort_results(results: pd.DataFrame) -> pd.DataFrame:
        return results.sort_values('worst_case', ascending=False)

    def _get_best_params(self, model, feature_set: str) -> dict:
        # Filter results for model and version
        results = self.results[np.logical_and(
            self.results['model'] == type(model).__name__,
            self.results['version'] == self.version,
        )]

        # Filter results for feature set & sort them
        results = self._sort_results(results[results['dataset'] == feature_set])

        # Warning for unoptimized results
        if 'Hyper Parameter' not in results['type'].values:
            warnings.warn('Hyper parameters not optimized for this combination')

        # Parse & return best parameters (regardless of if it's optimized)
        return Utils.io.parse_json(results.iloc[0]['params'])

    def _grid_search_iteration(self, model, parameter_set: str, feature_set: str):
        """
        INTERNAL | Grid search for defined model, parameter set and feature set.
        """
        print('\n[AutoML] Starting Hyper Parameter Optimization for {} on {} features ({} samples, {} features)'.format(
            type(model).__name__, feature_set, len(self.x), len(self.featureSets[feature_set])))

        # Cross-Validator
        cv_args = {'n_splits': self.cvSplits, 'shuffle': self.shuffle,
                   'random_state': 83847939 if self.shuffle else None}
        cv = KFold(**cv_args) if self.mode == 'regression' else StratifiedKFold(**cv_args)

        # Select right hyper parameter optimizer
        if self.gridSearchType == 'base':
            grid_search = BaseGridSearch(model, params=parameter_set, cv=cv, scoring=self.objective,
                                         candidates=self.gridSearchCandidates, timeout=self.gridSearchTimeout,
                                         verbose=self.verbose)
        elif self.gridSearchType == 'halving':
            grid_search = HalvingGridSearch(model, params=parameter_set, cv=cv, scoring=self.objective,
                                            candidates=self.gridSearchCandidates, verbose=self.verbose)
        elif self.gridSearchType == 'optuna':
            grid_search = OptunaGridSearch(model, timeout=self.gridSearchTimeout, cv=cv,
                                           candidates=self.gridSearchCandidates, scoring=self.objective,
                                           verbose=self.verbose)
        else:
            raise NotImplementedError('Only Base, Halving and Optuna are implemented.')
        # Get results
        results = grid_search.fit(self.x[self.featureSets[feature_set]], self.y)
        results = results.sort_values('worst_case', ascending=False)

        # Warn when best hyperparameters are close to predefined grid
        edge_params = grid_search.get_parameter_min_max()
        best_params = pd.Series(results['params'].iloc[0], name='best')
        params = edge_params.join(best_params, how='inner')

        def warn_when_too_close_to_edge(param: pd.Series, tol=0.01):
            # Min-max scaling
            scaled = np.array(param['best']) / (param['max'] - param['min'])
            # Check if too close and warn if so
            if not (tol < scaled < 1 - tol):
                warn_message = ('WARNING: Optimal value for parameter {} '
                                'is very close to edge case.\t({} = {})'
                                ).format(param.name, param.name, param['best'])
                warnings.warn(warn_message)

        params.apply(lambda p: warn_when_too_close_to_edge(p), axis=1)

        return results

    def _get_main_predictors(self, data):
        """
        Using Shapely Additive Explanations, this function calculates the main predictors for a given
        prediction and sets them into the class' memory.
        """
        # Shap is not implemented for all models.
        if type(self.bestModel).__name__ in ['SVC', 'BaggingClassifier', 'RidgeClassifier', 'LinearRegression', 'SVR',
                                             'BaggingRegressor']:
            features = self.settings['feature_processing']['featureImportance']['shap'][0]
            values = self.settings['feature_processing']['featureImportance']['shap'][1]
            self._main_predictors = {features[i]: values[i] for i in range(len(features))}

        else:
            if type(self.bestModel).__module__[:5] == 'Amplo':
                shap_values = np.array(TreeExplainer(self.bestModel.model).shap_values(data))
            else:
                shap_values = np.array(TreeExplainer(self.bestModel).shap_values(data))

            # Shape them (for multiclass it outputs ndim=3, for binary/regression ndim=2)
            if shap_values.ndim == 3:
                shap_values = shap_values[1]

            # Take mean over samples
            shap_values = np.mean(shap_values, axis=0)

            # Sort them
            inds = sorted(range(len(shap_values)), key=lambda x: -abs(shap_values[x]))

            # Set class attribute
            self._main_predictors = dict([(data.keys()[i], float(abs(shap_values[i]))) for i in inds])

        return self._main_predictors



import copy
import time
import optuna
import warnings
import numpy as np
import pandas as pd
from datetime import datetime
from typing import Dict, Union

from ._GridSearch import _GridSearch


class OptunaGridSearch(_GridSearch):

    def __init__(self, model, *args, **kwargs):
        """
        Wrapper for Optuna Grid Search. Takes any model supported by Amplo.AutoML.Modelling.
        The parameter search space is predefined for each model.

        Parameters
        ----------
        model obj: Model object to optimize
        cv obj: Scikit CV object
        scoring str: From Scikits Scorers*
        verbose int: How much to print
        timeout int: Time limit of optimization
        candidates int: Candidate limits to evaluate
        """
        super().__init__(model, *args, **kwargs)
        if 'params' in kwargs:
            warnings.warn('Parameter `params` has no effect')

    @staticmethod
    def _get_suggestion(trial: optuna.Trial, p_name: str, p_value: Union[tuple, list]) \
            -> Union[None, bool, int, float, str]:
        """Get suggestion for specific parameter

        Parameters
        ----------
        trial: optuna.Trial to sample from
        p_name: hyper parameter name
        p_value: contains necessary info (dtype, range, ...)

        Returns
        -------
        A specific parameter for given hyper parameter
        """

        # Read out
        p_type = p_value[0]  # parameter type (str)
        p_args = p_value[1]  # parameter arguments (list, tuple)
        p_kwargs = {}  # additional parameter keyword arguments, depends on parameter type

        # Sanity checks
        assert len(p_args) == 2 or p_type == 'categorical', \
            'Only categorical parameter can have more/less than two suggest args'

        # Find trial.suggest_{...} type
        if p_type == 'categorical':
            p_args = [p_args]  # enclose since will be expanded afterwards
            suggest = trial.suggest_categorical
        elif p_type == 'int':
            suggest = trial.suggest_int
        elif p_type == 'logint':
            p_kwargs += dict(log=True)
            suggest = trial.suggest_int
        elif p_type == 'uniform':
            suggest = trial.suggest_uniform
        elif p_type == 'loguniform':
            suggest = trial.suggest_loguniform
        else:
            raise NotImplementedError('Invalid parameter specification')

        # Suggest parameter given the arguments
        return suggest(p_name, *p_args)

    def _get_hyper_params(self, trial: optuna.Trial) -> Dict[str, Union[None, bool, int, float, str]]:
        """Use trial to sample from available grid search parameters

        Parameters
        ----------
        trial: Trial from Optuna study

        Returns
        -------
        Sampled grid search parameters
        """

        param_values = self._hyper_parameter_values
        conditionals = param_values.pop('CONDITIONALS', {})
        params = {}

        # Add non-conditional parameter suggestions (do not depend on presence of another)
        for name, value in param_values.items():
            params[name] = self._get_suggestion(trial, name, value)

        # Add conditional parameter suggestions that satisfy condition:
        for check_p_name, check_p_criteria in conditionals.items():
            for matching_value, additional_params in check_p_criteria:
                # Add parameters only if criterion is satisfied
                if matching_value != params.get(check_p_name, None):
                    # No match
                    continue
                for name, value in additional_params.items():
                    params[name] = self._get_suggestion(trial, name, value)

        return params

    def fit(self, x, y):
        if isinstance(y, pd.DataFrame):
            assert len(y.keys()) == 1, 'Multiple target columns not supported.'
            y = y[y.keys()[0]]
        assert isinstance(x, pd.DataFrame), 'X should be Pandas DataFrame'
        assert isinstance(y, pd.Series), 'Y should be Pandas Series or DataFrame'

        # Set mode
        self.binary = y.nunique() == 2
        self.samples = len(y)

        # Store
        self.x, self.y = x, y

        # Set up study
        study = optuna.create_study(sampler=optuna.samplers.TPESampler(seed=236868), direction='maximize')
        study.optimize(self.objective, timeout=self.timeout, n_trials=self.nTrials)

        # Parse results
        optuna_results = study.trials_dataframe()
        results = pd.DataFrame({
            'date': datetime.today().strftime('%d %b %y'),
            'model': type(self.model).__name__,
            'params': [x.params for x in study.get_trials()],
            'mean_objective': optuna_results['value'],
            'std_objective': optuna_results['user_attrs_std_value'],
            'worst_case': optuna_results['value'] - optuna_results['user_attrs_std_value'],
            'mean_time': optuna_results['user_attrs_mean_time'],
            'std_time': optuna_results['user_attrs_std_time']
        })

        return results

    def objective(self, trial):
        # Metrics
        scores = []
        times = []
        master = copy.deepcopy(self.model)

        # Cross Validation
        for t, v in self.cv.split(self.x, self.y):
            # Split data
            xt, xv, yt, yv = self.x.iloc[t], self.x.iloc[v], self.y.iloc[t], self.y.iloc[v]

            # Train model
            t_start = time.time()
            model = copy.deepcopy(master)
            model.set_params(**self._get_hyper_params(trial))
            model.fit(xt, yt)

            # Results
            scores.append(self.scoring(model, xv, yv))
            times.append(time.time() - t_start)

        # Set manual metrics
        trial.set_user_attr('mean_time', np.mean(times))
        trial.set_user_attr('std_time', np.std(times))
        trial.set_user_attr('std_value', np.std(scores))

        # Stop trail (avoid overwriting)
        if trial.number == self.nTrials:
            trial.study.stop()

        return np.mean(scores)



import logging
import warnings


__all__ = ['get_model', 'hist_search']


def get_model(model_str, **args):
    # Import here to prevent ImportError (due to circular import)
    from Amplo.AutoML.Modeller import Modeller
    try:
        model = [mod for mod in Modeller(**args).return_models()
                 if type(mod).__name__ == model_str]
        return model[0]
    except IndexError:
        raise IndexError('Model not found.')


def hist_search(array, value):
    """
    Binary search that finds the index in order to fulfill
    ``array[index] <= value < array[index + 1]``

    Parameters
    ----------
    array : array of float
    value : float

    Returns
    -------
    int
        Bin index of the value
    """

    # Return -1 when no bin exists
    if value < array[0] or value >= array[-1]:
        logging.debug(f'No bin (index) found for value {value}. Array(Min: {array[0]}, Max: {array[-1]})')
        return -1

    # Initialize min and max bin index
    low = 0
    high = len(array) - 1

    # Bin search
    countdown = 30
    while countdown > 0:
        # Count down
        countdown -= 1

        # Set middle bin index
        middle = low + (high - low) // 2

        if low == middle == high - 1:  # stop criterion
            return middle

        if value < array[middle]:  # array[low] <= value < array[middle]
            high = middle
        elif value >= array[middle]:  # array[middle] <= value < array[high]
            low = middle

    warnings.warn(RuntimeWarning('Operation took too long. Returning -1 (no match).'))
    return -1



import pytest
import os

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification, make_regression

from Amplo import Pipeline
from Amplo.AutoML import Modeller


@pytest.fixture
def make_data(mode):
    if mode == 'classification':
        x, y = make_classification()
    elif mode == 'regression':
        x, y = make_regression()
    else:
        raise ValueError('Invalid mode')
    yield pd.DataFrame(x), pd.Series(y)


class TestPipeline:

    @pytest.mark.parametrize('mode', ['classification', 'regression'])
    @pytest.mark.parametrize('n_samples', [100, 100_000])
    def test_main_predictors(self, mode, make_data, n_samples):
        # Test mode
        x, y = make_data
        pipeline = Pipeline(grid_search_iterations=1, grid_search_candidates=1, plot_eda=False)
        pipeline.fit(x, y)
        for model in Modeller(mode=mode, samples=n_samples).return_models():
            print(model)
            x_c, _ = pipeline.convert_data(x)
            model.fit(x_c, y)
            pipeline.bestModel = model
            pipeline.predict(x)
            assert isinstance(pipeline._main_predictors, dict), 'Main predictors not dictionary.'

    def test_no_dirs(self):
        pipeline = Pipeline(no_dirs=True)
        assert not os.path.exists('AutoML'), 'Directory created'

    def test_no_args(self):
        x, y = make_regression()
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(x, y)

    @pytest.mark.parametrize('mode', ['classification', 'regression'])
    def test_mode_detector(self, mode, make_data):
        x, y = make_data
        pipeline = Pipeline()
        pipeline._read_data(x, y)._mode_detector()
        assert pipeline.mode == mode

    def test_create_folders(self):
        x, y = make_classification()
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(x, y)

        # Test Directories
        assert os.path.exists('AutoML')
        assert os.path.exists('AutoML/Data')
        assert os.path.exists('AutoML/Features')
        assert os.path.exists('AutoML/Production')
        assert os.path.exists('AutoML/Documentation')
        assert os.path.exists('AutoML/Results.csv')

    def test_read_write_csv(self):
        """
        Check whether intermediate data is stored and read correctly
        """
        # Set path
        data_path = 'test_data.csv'

        # Test single index
        data_write = pd.DataFrame(np.random.randint(0, 100, size=(10, 10)),
                                  columns=[f'feature_{i}' for i in range(10)])
        data_write.index.name = 'index'
        Pipeline._write_csv(data_write, data_path)
        data_read = Pipeline._read_csv(data_path)
        assert data_write.equals(data_read), 'Read data should be equal to original data'

        # Test multi-index (cf. IntervalAnalyser)
        data_write = data_write.set_index(data_write.columns[-2:].to_list())
        data_write.index.names = ['log', 'index']
        Pipeline._write_csv(data_write, data_path)
        data_read = Pipeline._read_csv(data_path)
        assert data_write.equals(data_read), 'Read data should be equal to original data'

        # Remove data
        os.remove(data_path)



from abc import abstractmethod
from typing import Any, Dict, List, Tuple, Optional, Union
import re

import numpy as np
import pandas as pd
from sklearn.model_selection import KFold
from sklearn.metrics import SCORERS

import multiprocessing as mp


class _GridSearch:
    def __init__(
            self,
            model,
            params=None,
            candidates=250,
            timeout=None,
            cv=KFold(n_splits=10),
            scoring='accuracy',
            verbose=0,
    ):
        # Input tests
        assert model is not None, 'Need to provide a model'
        if hasattr(model, 'is_fitted'):
            assert not model.is_fitted(), 'Model already fitted'
        if scoring is None:
            if 'Classifier' in type(model).__name__:
                self.scoring = SCORERS['accuracy']
            elif 'Regressor' in type(model).__name__:
                self.scoring = SCORERS['neg_mean_squared_error']
            else:
                raise ValueError('Model mode unknown')

        # Set class attributes
        self.model = model
        self.params = params
        self.nTrials = candidates
        self.timeout = timeout
        self.cv = cv
        self.scoring = SCORERS[scoring] if isinstance(scoring, str) else scoring
        self.verbose = verbose

        self.x, self.y = None, None
        self.binary = True
        self.samples = None

        # Model specific settings
        if type(self.model).__name__ == 'LinearRegression':
            self.nTrials = 1

    @property
    def _hyper_parameter_values(self) \
            -> Dict[str, Tuple[str, List[Union[str, float]], Optional[int]]]:
        """Get model specific hyper parameter values, indicating predefined
        search areas to optimize.

        Notes
        -----

        Each item of the output dictionary consists of:
            - parameter name (str), and
            - parameter specifications (tuple)
                - parameter type (str)
                - parameter arguments (list)
                - number of distinct values (int, optional) [only for exhaustive grid search]

        Parameter types include:
            - 'categorical': categorical values
            - 'int': discretized uniform value space
            - 'logint': discretized logarithmic uniform value space
            - 'uniform': uniform value space
            - 'loguniform': logarithmic uniform value space

        Parameter arguments are:
            - [categorical]: a list of all options
            - [int, logint, uniform, loguniform]: a tuple with min and max value

        **Special case (conditionals):**
            In some cases, one wants to grid-search certain parameters only
            if another parameter condition is present. Such conditions are
            specified via the dedicated key 'CONDITIONALS'.
        """

        # Extract model name & type
        model_name = type(self.model).__name__
        model_type = re.split(r'Regressor|Classifier', model_name)[0]
        # Determine whether it's classification or regression
        is_regression = bool(re.match(r'.*(Regression|Regressor|SVR)', model_name))
        is_classification = bool(re.match(r'.*(Classification|Classifier|SVC)', model_name))
        assert is_regression or is_classification,\
            'Could not determine mode (regression or classification)'

        # Define min-max-function
        def minimax(min_, value, max_):
            return max(min_, min(value, max_))

        # Find matching model and return its parameter values

        if model_name == 'LinearRegression':
            return {}

        elif model_name == 'Lasso' or 'Ridge' in model_name:
            return dict(
                alpha=('uniform', [0, 10], 25),
            )

        elif model_name in ('SVR', 'SVC'):
            return dict(
                gamma=('categorical', ['scale', 'auto', 0.001, 0.01, 0.1, 0.5, 1]),
                C=('uniform', [0.001, 10], 25),
            )

        elif model_type == 'KNeighbors':
            return dict(
                n_neighbors=('int', [5, minimax(5, self.samples // 10, 50)], 5),
                weights=('categorical', ['uniform', 'distance']),
                leaf_size=('int', [1, minimax(1, self.samples // 10, 100)], 5),
                n_jobs=('categorical', [mp.cpu_count() - 1]),
            )

        elif model_type == 'MLP':
            raise NotImplementedError('MLP is not supported')

        elif model_type == 'SGD':
            params = dict(
                loss=('categorical', ['squared_loss', 'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive']),
                penalty=('categorical', ['l2', 'l1', 'elasticnet']),
                alpha=('uniform', [0, 10], 5),
                max_iter=('int', [250, 1000], 3),
            )
            if is_classification:
                params.update(loss=('categorical', ['hinge', 'log', 'modified_huber', 'squared_hinge']))
            return params

        elif model_type == 'DecisionTree':
            params = dict(
                criterion=('categorical', ['squared_error', 'friedman_mse', 'absolute_error', 'poisson']),
                max_depth=('int', [3, minimax(3, int(np.log2(self.samples)), 25)], 4),
            )
            if is_classification:
                params.update(criterion=('categorical', ['gini', 'entropy']))
            return params

        elif model_type == 'AdaBoost':
            params = dict(
                n_estimators=('int', [25, 250], 5),
                loss=('categorical', ['linear', 'square', 'exponential']),
                learning_rate=('loguniform', [0.001, 1], 10),
            )
            if is_classification:
                params.pop('loss', None)
            return params

        elif model_type == 'Bagging':
            params = dict(
                n_estimators=('int', [10, 250], 4),
                max_samples=('uniform', [0.5, 1.0], 4),
                max_features=('uniform', [0.5, 1.0], 4),
                n_jobs=('categorical', [mp.cpu_count() - 1]),
            )
            return params

        elif model_type == 'CatBoost':
            params = dict(
                n_estimators=('int', [500, 2000], 5),
                verbose=('categorical', [0]),
                od_pval=('categorical', [1e-5]),
                loss_function=('categorical', ['MAE', 'RMSE']),
                learning_rate=('loguniform', [0.001, 0.5], 5),
                l2_leaf_reg=('uniform', [0, 10], 5),
                depth=('int', [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                min_data_in_leaf=('int', [1, minimax(1, self.samples // 10, 1000)], 5),
                grow_policy=('categorical', ['SymmetricTree', 'Depthwise', 'Lossguide']),
            )
            if is_classification:
                params.update(loss_function=('categorical', ['Logloss' if self.binary else 'MultiClass']))
            return params

        elif model_type == 'GradientBoosting':
            params = dict(
                loss=('categorical', ['ls', 'lad', 'huber']),
                learning_rate=('loguniform', [0.001, 0.5], 10),
                max_depth=('int', [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                n_estimators=('int', [100, 1000], 4),
                min_samples_leaf=('int', [1, minimax(1, int(self.samples / 10), 1000)], 3),
                max_features=('uniform', [0.5, 1], 3),
                subsample=('uniform', [0.5, 1], 3)
            )
            if is_classification:
                params.update(loss=('categorical', ['deviance', 'exponential']))
            return params

        elif model_type == 'HistGradientBoosting':
            params = dict(
                loss=('categorical', ['least_squares', 'least_absolute_deviation']),
                learning_rate=('loguniform', [0.001, 0.5], 10),
                max_iter=('int', [100, 250], 4),
                max_leaf_nodes=('int', [30, 150], 4),
                max_depth=('int', [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                min_samples_leaf=('int', [1, minimax(1, int(self.samples / 10), 1000)], 4),
                l2_regularization=('uniform', [0, 10], 5),
                max_bins=('int', [100, 255], 4),
                early_stopping=('categorical', [True])
            )
            if is_classification:
                params.pop('loss', None)
            return params

        elif model_type == 'RandomForest':
            params = dict(
                n_estimators=('int', [50, 1000], 5),
                criterion=('categorical', ['squared_error', 'absolute_error']),
                max_depth=('int', [3, minimax(3, int(np.log2(self.samples)), 15)], 4),
                max_features=('categorical', ['auto', 'sqrt']),
                min_samples_split=('int', [2, 50], 4),
                min_samples_leaf=('int', [1, minimax(1, self.samples // 10, 1000)], 5),
                bootstrap=('categorical', [True, False]),
            )
            if is_classification:
                params.update(criterion=('categorical', ['gini', 'entropy']))
            return params

        elif model_type == 'XGB':
            params = dict(
                objective=('categorical', ['reg:squarederror']),
                eval_metric=('categorical', ['rmse']),
                booster=('categorical', ['gbtree', 'gblinear', 'dart']),
                alpha=('loguniform', [1e-8, 1], 10),
                learning_rate=('loguniform', [0.001, 0.5], 5),
                n_jobs=('categorical', [mp.cpu_count() - 1]),
            )
            params['lambda'] = ('loguniform', [1e-8, 1], 10)
            if is_classification:
                params.update(
                    objective=('categorical', ['multi:softprob']),
                    eval_metric=('categorical', ['logloss']),
                )
            params['CONDITIONALS'] = dict(
                booster=[
                    ('gbtree', dict(
                        max_depth=('int', [1, minimax(1, int(np.log2(self.samples)), 10)], 5),
                        eta=('loguniform', [1e-8, 1], 5),
                        gamma=('loguniform', [1e-8, 1], 5),
                        grow_policy=('categorical', ['depthwise', 'lossguide']),
                    )),
                    ('dart', dict(
                        max_depth=('int', [1, minimax(1, int(np.log2(self.samples)), 10)], 5),
                        eta=('loguniform', [1e-8, 1], 5),
                        gamma=('loguniform', [1e-8, 1], 5),
                        grow_policy=('categorical', ['depthwise', 'lossguide']),
                        sample_type=('categorical', ['uniform', 'weighted']),
                        normalize_type=('categorical', ['tree', 'forest']),
                        rate_drop=('loguniform', [1e-8, 1], 5),
                        skip_drop=('loguniform', [1e-8, 1], 5),
                    )),
                ],
            )
            return params

        elif model_type == 'LGBM':
            if is_regression:
                return dict(
                    num_leaves=('int', [10, 150], 5),
                    min_data_in_leaf=('int', [1, minimax(1, self.samples // 10, 1000)], 0),
                    min_sum_hessian_in_leaf=('uniform', [0.001, 0.5], 0),
                    colsample_bytree=('uniform', [0, 1], 5),
                    reg_alpha=('uniform', [0, 1], 5),
                    reg_lambda=('uniform', [0, 1], 5),
                    verbosity=('categorical', [-1]),
                    n_jobs=('categorical', [mp.cpu_count() - 1]),
                )
            else:  # is_classification
                return dict(
                    objective=('categorical', ['binary' if self.binary else 'multiclass']),
                    metric=('categorical',
                            ['binary_error', 'auc', 'average_precision', 'binary_logloss']
                            if self.binary else ['multi_error', 'multi_logloss', 'auc_mu']),
                    boosting_type=('categorical', ['gbdt']),
                    lambda_l1=('loguniform', [1e-8, 10], 4),
                    lambda_l2=('loguniform', [1e-8, 10], 4),
                    num_leaves=('int', [10, 5000], 4),
                    max_depth=('int', [5, 20], 4),
                    min_data_in_leaf=('int', [1, minimax(1, self.samples // 10, 1000)], 0),
                    min_gain_to_split=('uniform', [0, 5], 0),
                    feature_fraction=('uniform', [0.4, 1], 0),
                    bagging_fraction=('uniform', [0.4, 1], 0),
                    bagging_freq=('int', [1, 7], 0),
                    verbosity=('categorical', [-1]),
                    n_jobs=('categorical', [mp.cpu_count() - 1]),
                )

        # Raise error if no match was found
        raise NotImplementedError('Hyper parameter tuning not implemented for {}'.format(model_name))

    def get_parameter_min_max(self) -> pd.DataFrame:
        """Get all min and max values from model-specific set of parameters.
        Omit categorical parameters as min and max values are ambiguous.

        Returns
        -------
        param_min_max (pd.DataFrame)
        """

        # Get all model's parameters
        param_values = self._hyper_parameter_values
        # Pop conditionals and integrate all into `param_values`
        conditionals = param_values.pop('CONDITIONALS', {})
        for check_p_name, check_p_criteria in conditionals.items():
            for matching_value, additional_params in check_p_criteria:
                for name, value in additional_params.items():
                    param_values[name] = value

        # Filter for min and max in non-categorical parameters
        param_min_max = {}
        for p_name, value in param_values.items():
            p_type = value[0]
            p_args = value[1]
            if p_type in ('int', 'logint', 'uniform', 'loguniform'):
                # Sanity check
                assert len(p_args) == 2, 'A {} should have a min and a max value'.format(p_type)
                # Add item to dict
                add_item = {p_name: {'min': p_args[0], 'max': p_args[1]}}
                param_min_max.update(add_item)

        # Combine all values to pd.DataFrame
        param_min_max = pd.DataFrame(param_min_max).T
        return param_min_max

    @abstractmethod
    def _get_hyper_params(self, *args, **kwargs) -> Dict[str, Any]:
        """Translate `self._hyper_parameter_values` to grid-search specific
        distributions or samples.

        Parameters
        ----------
        *args (optional): grid-search specific arguments
        **kwargs (optional): grid-search specific arguments

        Returns
        -------
        Grid search parameters
        """
        pass

    @abstractmethod
    def fit(self, x, y) -> pd.DataFrame:
        """Run fit with model-specific set of parameters

        Parameters
        ----------
        x (array-type): data features
        y (array-type): data labels

        Returns
        -------
        results (pd.DataFrame)
        """
        pass



import os
import json
import warnings
from typing import Union, Tuple
from pathlib import Path

import pandas as pd


__all__ = ['boolean_input', 'parse_json', 'read_pandas', 'merge_logs']


def boolean_input(question: str) -> bool:
    x = input(question + ' [y / n]')
    if x.lower() == 'n' or x.lower() == 'no':
        return False
    elif x.lower() == 'y' or x.lower() == 'yes':
        return True
    else:
        print('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def parse_json(json_string: Union[str, dict]) -> Union[str, dict]:
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(json_string
                              .replace("'", '"')
                              .replace("True", "true")
                              .replace("False", "false")
                              .replace("nan", "NaN")
                              .replace("None", "null"))
        except json.decoder.JSONDecodeError:
            print('[AutoML] Cannot validate, impassable JSON.')
            print(json_string)
            return json_string


def read_pandas(path: Union[str, Path]) -> pd.DataFrame:
    """
    Wrapper for various read functions

    Returns
    -------
    pd.DataFrame
    """
    f_ext = Path(path).suffix
    if f_ext == '.csv':
        return pd.read_csv(path)
    elif f_ext == '.json':
        return pd.read_json(path)
    elif f_ext == '.xml':
        return pd.read_xml(path)
    elif f_ext == '.feather':
        return pd.read_feather(path)
    elif f_ext == '.parquet':
        return pd.read_parquet(path)
    elif f_ext == '.stata':
        return pd.read_stata(path)
    elif f_ext == '.pickle':
        return pd.read_pickle(path)
    else:
        raise NotImplementedError('File format not supported.')


def merge_logs(path_to_folder: Union[str, Path], target: str = 'labels') -> Tuple[pd.DataFrame, pd.DataFrame]:
    r"""
    Combine log files from given directory into a multi-indexed dataframe

    Notes
    -----
    Make sure that each protocol is located in a sub folder whose name represents the respective label.

    A directory structure example:
        |   ``path_to_folder``
        |   ``├─ Label_1``
        |   ``│   ├─ Log_1.*``
        |   ``│   └─ Log_2.*``
        |   ``├─ Label_2``
        |   ``│   └─ Log_3.*``
        |   ``└─ ...``

    Parameters
    ----------
    path_to_folder : str or Path
        Parent directory
    target : str
        Column name for target

    Returns
    -------
    data : pd.DataFrame
        All logs concatenated into one multi-indexed dataframe.
        Multi-index names are ``log`` and ``index``.
        Target column depicts the folder name.
    metadata : pd.DataFrame
        Metadata with ``folder``, ``file``, ``full_path`` and ``last_modified`` column, single-indexed
    """
    # Tests
    assert Path(path_to_folder).is_dir(), 'Invalid path to directory'
    assert Path(path_to_folder).exists(), 'Directory does not exist'
    assert isinstance(target, str), 'Target name should be a string'
    assert target != '', 'Do not use an empty string as target name'

    # Result init
    data = []
    metadata = []
    n_files = 0

    # Loop through folders
    for folder in sorted(Path(path_to_folder).iterdir()):

        # Loop through files (ignore hidden files)
        for file in sorted(folder.glob('[!.]*.*')):

            # Read df
            try:
                datum = read_pandas(file)
            except pd.errors.EmptyDataError:
                # Skip when empty
                warnings.warn(f'[AutoML] Empty file detected: {file}')
                continue

            # Set labels
            datum[target] = folder.name

            # Set index
            datum.set_index(pd.MultiIndex.from_product([[n_files], datum.index.values], names=['log', 'index']),
                            inplace=True)

            # Set metadata
            meta = pd.DataFrame({
                'folder': [folder.name], 'file': [file.name], 'full_path': [str(file.resolve())],
                'last_modified': [os.path.getmtime(str(file))]
            }, index=pd.Index([n_files], name='log'))

            # Add to list
            data.append(datum)
            metadata.append(meta)

            # Increment
            n_files += 1

    if n_files == 0:
        raise FileNotFoundError('Folder directory seems to be empty. Check whether you specified the correct path')
    elif n_files == 1:
        # Omit concatenation when only one item
        return data[0], metadata[0]
    else:
        # Concatenate dataframes
        return pd.concat(data), pd.concat(metadata)



from pathlib import Path
import logging
import colorlog


__all__ = ['get_logger']


_nameToLevel = {
    'CRITICAL': logging.CRITICAL,
    'FATAL': logging.FATAL,
    'ERROR': logging.ERROR,
    'WARN': logging.WARNING,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'NOTSET': logging.NOTSET,
}


def _check_logging_level(level):
    """
    Check input whether it is a valid logging level

    Parameters
    ----------
    level : int or str
        Logging level

    Returns
    -------
    int
        (valid) logging level

    Raises
    ------
    ValueError
        If the given logging level is not valid
    """
    # Inspired by logging/__init__.py
    if isinstance(level, int):
        if level not in set(_nameToLevel.values()):
            raise ValueError("Unknown level: %r" % level)
        rv = level
    elif str(level) == level:
        if level not in _nameToLevel:
            raise ValueError("Unknown level: %r" % level)
        rv = _nameToLevel[level]
    else:
        raise TypeError("Level not an integer or a valid string: %r" % level)
    return rv


def get_logger(name='AutoML', log_path=None, level=logging.NOTSET, capture_warnings=True) -> logging.Logger:
    """
    Create a logging Logger

    Parameters
    ----------
    name : str
        Name of the logger
    log_path : str or Path, optional
        Specifies the logging path, in case you want to store the logs in a file.
    level : int or str, optional
        Logging level
    capture_warnings : bool
        Whether to capture `warnings.warn(...)` with the logger.
        Note that this option is set globally!

    Returns
    -------
    logging.Logger
    """
    # TODO: Add option `parent` to use `parent.getChild()`
    #  -> This is part of https://amplo.atlassian.net/browse/AML-71

    # Safety check for logging level
    level = _check_logging_level(level)

    # Get custom logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Set console handler
    console_formatter = colorlog.ColoredFormatter(
        '%(white)s[%(name)s] %(log_color)s%(levelname)s: %(message)s %(white)s<%(filename)s:%(lineno)d>')
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    # Set file handler
    if log_path is not None:
        if Path(log_path).suffix != '.log':
            logger.warning('It is recommended naming a log file as `*.log`')
        file_formatter = logging.Formatter('[%(name)s] %(levelname)s: %(message)s')
        file_handler = logging.FileHandler(log_path)
        file_handler.setLevel(level)
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)

    # Capture warnings
    if capture_warnings:
        # Capture warnings from `warnings.warn(...)`
        logging.captureWarnings(True)
        # Get py-warnings logger and add handler
        py_warnings_logger = logging.getLogger('py.warnings')
        py_warnings_logger.addHandler(console_handler)

    return logger



import pytest
import os
import json
import joblib
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification, make_regression
from sklearn.metrics import r2_score, log_loss

from Amplo import Pipeline
from Amplo.Utils.testing import make_interval_data

from tests import rmtree


def make_data(mode, target='target'):
    if mode == 'regression':
        x, y = make_regression()
    elif mode == 'classification':
        x, y = make_classification()
    else:
        raise ValueError('Not implemented')
    data = pd.DataFrame(x, columns=['Feature_{}'.format(i) for i in range(x.shape[1])])
    data[target] = y
    return data


@pytest.fixture(scope='class', params=['regression', 'classification'])
def make_mode(request):
    mode = request.param
    target = 'target'
    request.cls.mode = mode
    request.cls.target = target
    request.cls.data = make_data(mode, target)
    yield


@pytest.mark.usefixtures('make_mode')
class TestPipelineByMode:

    def test_mode_detector(self):
        # Numerical test
        pipeline = Pipeline(no_dirs=True, target=self.target)
        pipeline._read_data(self.data)._mode_detector()
        assert pipeline.mode == self.mode

        # Categorical test
        if self.mode == 'classification':
            # Convert to categorical
            data = self.data.copy()
            data[self.target] = [f'Class_{v}' for v in data[self.target].values]
            # Detect mode
            pipeline = Pipeline(no_dirs=True, target=self.target)
            pipeline._read_data(data)._mode_detector()
            assert pipeline.mode == self.mode

    def test_stacking(self):
        # Set stacking model
        if self.mode == 'classification':
            stacking_model = 'StackingClassifier'
        else:
            stacking_model = 'StackingRegressor'
        # Fit pipeline
        pipeline = Pipeline(
            target=self.target, grid_search_type=None,
            stacking=True, feature_timeout=5)
        pipeline.fit(self.data, model=stacking_model)

    def test_dir_and_settings(self):
        pipeline = Pipeline(
            target=self.target,
            name='Auto' + str(self.mode).capitalize(),
            mode=self.mode,
            objective='r2' if self.mode == 'regression' else 'neg_log_loss',
            feature_timeout=5,
            grid_search_iterations=0,
            plot_eda=False,
            process_data=False,
            document_results=False)
        pipeline.fit(self.data)

        # Test Directories
        assert os.path.exists('AutoML')
        assert os.path.exists('AutoML/Data')
        assert os.path.exists('AutoML/Features')
        assert os.path.exists('AutoML/Production')
        assert os.path.exists('AutoML/Documentation')
        assert os.path.exists('AutoML/Results.csv')

        if self.mode == 'classification':
            # Pipeline Prediction
            prediction = pipeline.predict_proba(self.data)
            assert log_loss(self.data[self.target], prediction) > -1

        elif self.mode == 'regression':
            # Test data handling
            c, _ = pipeline.convert_data(self.data.drop(self.target, axis=1))
            x = pipeline.x[pipeline.settings['features']]
            y = self.data[self.target]
            assert np.allclose(c.values, x.values), f'Inconsistent X: max diff: {np.max(abs(c.values - x.values)):.2e}'
            assert np.allclose(y, pipeline.y), 'Inconsistent y-data'

            # Pipeline Prediction
            prediction = pipeline.predict(self.data)
            assert len(prediction.shape) == 1
            assert r2_score(self.data[self.target], prediction) > 0.75

        else:
            raise ValueError(f'Invalid mode {self.mode}')

        # Settings prediction
        settings = json.load(open('AutoML/Production/v1/Settings.json', 'r'))
        model = joblib.load('AutoML/Production/v1/Model.joblib')
        p = Pipeline(no_dirs=True)
        p.load_settings(settings)
        p.load_model(model)
        if self.mode == 'classification':
            assert np.allclose(p.predict_proba(self.data), prediction)
        else:
            assert np.allclose(p.predict(self.data), prediction)

        # Check settings
        assert 'pipeline' in settings
        assert 'version' in settings
        assert 'name' in settings['pipeline']
        assert 'model' in settings
        assert 'amplo_version' in settings
        assert 'params' in settings
        assert 'drift_detector' in settings
        assert 'features' in settings
        assert 'validation' in settings
        assert 'data_processing' in settings


class TestPipeline:

    def test_drift(self):
        # todo connect with logger
        pytest.skip('This test is not yet implemented')

    def test_all_models(self):
        # TODO
        pytest.skip('This test is not yet implemented')

    def test_interval_analyzer(self):
        """Use interval analyzer by default when a folder with logs is presented to pipeline"""
        # Create dummy data
        data_dir = './test_data'
        make_interval_data(directory=data_dir)
        # TODO: does it have to work also with n_labels=1 and n_logs=1?

        # Pipeline
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(data_dir)

        # Check if IntervalAnalyser is fitted
        assert pipeline.intervalAnalyser.is_fitted, 'IntervalAnalyser was not fitted'

        # Check data handling
        assert Path('AutoML/Data/Interval_Analyzed_v1.csv').exists(), \
            'Interval-analyzed data was not properly stored'
        assert list(pipeline.data.index.names) == ['log', 'index'], 'Index is incorrect'
        # TODO: add checks for settings

        # Remove dummy data
        rmtree(data_dir, must_exist=True)



import pytest
import itertools
import random

from Amplo.API import API
from Amplo.Utils.testing import make_interval_data, make_production_data

from tests.unit.api import TestAPI


class TestInterface(TestAPI):

    def test_iter_data(self):
        # Create dummy dataset
        paths_orig, infos_orig = [], []
        rng = random.Random(354687657938)
        for t, m, s in itertools.product(range(3), repeat=3):
            if rng.choice([True, False]):
                # Create "gaps" in data / leave some out
                continue
            team, machine, service = f'Team{t}', f'Machine{m}', f'Service{s}'
            paths_orig += [(self.sync_dir / f'{team}/{machine}/{service}').resolve()]
            infos_orig += [dict(team=team, machine=machine, service=service)]
            make_interval_data(directory=paths_orig[-1] / 'data')

        # Define helper function
        def is_selected(orig_lookup: dict, check_lookup: dict, name: str):
            """Return True if a selection match was found"""
            orig = orig_lookup[f'{name}']
            check = check_lookup[f'{name}s']
            if check is None or check == '*' or orig in check:
                return True
            else:
                return False

        # Define test cases
        select_kwargs = [
            # Get all combinations
            dict(teams=None, machines=None, services=None),
            # Single selection
            dict(teams='Team0', machines='Machine0', services='Service0'),
            # Mixed selection
            dict(teams=None, machines='Machine0', services=['Service0', 'Service1']),
        ]
        for kwargs in select_kwargs:
            # Get all paths from ``_iter_data``
            paths_actual, infos_actual = [], []
            for path, team, machine, service in API(self.sync_dir, **kwargs)._iterate_data(level='service'):
                paths_actual += [path]
                infos_actual += [dict(team=team, machine=machine, service=service)]

            # Define subset
            paths_target, infos_target = [], []
            for path, info in zip(paths_orig, infos_orig):
                if any(not is_selected(info, kwargs, key) for key in info.keys()):
                    continue
                paths_target += [path]
                infos_target += [info]

            # Assertions
            assert set(paths_actual) == set(paths_target), \
                f'Path check has failed with kwargs: {kwargs}'
            assert set(str(i) for i in infos_actual) == set(str(i) for i in infos_target), \
                f'Info check has failed with kwargs: {kwargs}'

    def test_download(self):
        # TODO: Set up a data selection that is quick to download
        # select = dict(teams='Amplo', machines='TestMachine', services='Unlabelled')
        select = dict(teams='Demo', machines='Charger 75kW', services='Diagnostics', issues='Blocked Radiator')
        # Set up API and synch data
        api = API(self.sync_dir, **select, verbose=2)
        api.sync_data()
        # Assert existence of folder
        assert (self.sync_dir / select['teams'] / select['machines'] / select['services']
                / 'data' / select['issues']).exists(), 'Download has failed or is not in assumed directory'

    def test_training(self):
        # Create a small dummy dataset
        dummy_data_path = self.sync_dir / 'DummyTeam' / 'DummyMachine' / 'DummyService' / 'data'
        make_interval_data(n_labels=3, directory=dummy_data_path)
        # Set up api and train models
        api = API(self.sync_dir, verbose=2)
        api.train_models(grid_search_iterations=0)
        # Assertions
        assert len(api._trained_model_args) > 0, \
            'No model has been trained or failed to append model arguments to list'

    def test_upload(self):
        """
        This only tests that upload_latest_model doesn't error.
        - Good to check whether the file exists in the cloud.
        - Also needs some cleanup to make sure it doesn't stay there.
        - Also does upload_latest_model truly take the latest?
        - What if latest locally and in the cloud are outdated? It needs to go in the right folder.
        """
        # TODO: Implement as soon as platform is out of beta mode
        pytest.skip('Test not yet implemented')

        # Make dummy production data
        issue_dir, kwargs = make_production_data(
            self.sync_dir, team='Demo', machine='Charger 75kW', service='Diagnostics')
        # Set up api and upload model
        api = API(self.sync_dir, verbose=2)
        api.upload_models([[issue_dir, kwargs['team'], kwargs['machine'], kwargs['service'],
                            kwargs['issue'], kwargs['version']]])



import warnings
from pathlib import Path
from typing import Union

import faiss
import numpy as np
import pandas as pd

from Amplo.Utils.io import merge_logs
from Amplo.Utils.data import check_dataframe_quality, check_pearson_correlation


class IntervalAnalyser:

    noise = 'Noise'

    def __init__(self,
                 target: str = None,
                 norm: str = 'euclidean',  # TODO: implement functionality
                 min_length: int = 1000,
                 n_neighbors: int = None,
                 n_trees: int = 10,
                 verbose: int = 1,
                 ):
        """
        Interval Analyser for Log file classification. Has two purposes:
        - Remove healthy data in longer, faulty logs
        - Remove redundant data in large datasets

        Uses Facebook's FAISS for K-Nearest Neighbors approximation.

        ** IMPORTANT **
        To use this interval analyser, make sure that your logs are located in a folder of their label, with one parent
        folder with all labels, e.g.:
        +-- Parent Folder
        |   +-- Label_1
        |       +-- Log_1.*
        |       +-- Log_2.*
        |   +-- Label_2
        |       +-- Log_3.*

        Parameters
        ----------
        target : str
            Target column name
        norm : str
            Optimization metric for K-Nearest Neighbors
            NOTE: This option has no effect, yet!
        min_length : int
            Minimum length to cut off, everything shorter is left untouched
        n_neighbors : int, optional
            Quantity of neighbors, default to 3 * log length
        n_trees : int
            Quantity of trees
        """
        # Test
        assert norm in ['euclidean', 'manhattan', 'angular', 'hamming', 'dot']

        # Parameters
        self.target = target or 'labels'
        self.min_length = min_length
        self.norm = norm
        self.n_trees = n_trees
        self.n_neighbors = n_neighbors
        self.verbose = verbose

        # Initializers
        self._labels = None
        self._features = None
        self._metadata = None  # optional metadata from `merge_logs`
        # self._mins = pd.DataFrame(index=[0])
        # self._maxs = pd.DataFrame(index=[0])
        self._engine = None
        self._distributions = None
        self._noise_indices = None
        self.avg_samples = None
        self.n_samples = None
        self.n_columns = None
        self.n_files = None
        self.n_folders = None

        # Flags
        self.is_fitted = False

    @property
    def orig_data(self) -> pd.DataFrame:
        """
        Returns
        -------
        pd.DataFrame
            Original data containing both, feature and label columns
        """
        return pd.concat([self._features, self._labels], axis=1)

    @property
    def data_with_noise(self) -> pd.DataFrame:
        """
        Returns
        -------
        pd.DataFrame
            Original data where noise is labelled as ``self.noise``

        Notes
        -----
        Depending on the dtype of the noise attribute, the dtype of the labels' column will be affected.
        """
        data = self.orig_data
        data.loc[self._noise_indices, self.target] = self.noise
        return data

    @property
    def data_without_noise(self) -> pd.DataFrame:
        """
        Returns
        -------
        pd.DataFrame
            Original data but without noise rows
        """
        return self.orig_data.drop(self._noise_indices)

    def fit_transform(self, data_or_path: Union[pd.DataFrame, str, Path], labels: pd.Series = None) -> pd.DataFrame:
        """
        Function that runs the K-Nearest Neighbors and returns a dataframe with the sensitivities.

        Parameters
        ----------
        data_or_path : pd.DataFrame or str or Path
            Multi-indexed dataset containing feature (and target) columns
            or path to folder that is correctly structured (see Notes)
        labels : pd.Series, optional
            Multi-indexed dataset containing target columns.

        Notes
        -----
        In case you provide a path-like argument to ``features_or_path``,
        make sure that each protocol is located in a sub folder whose name represents the respective label.

        A directory structure example:
            |   ``path_to_folder``
            |   ``├─ Label_1``
            |   ``│   ├─ Log_1.*``
            |   ``│   └─ Log_2.*``
            |   ``├─ Label_2``
            |   ``│   └─ Log_3.*``
            |   ``└─ ...``

        Returns
        -------
        pd.DataFrame
            Estimation of correlation strength
        """

        # Parse data
        features, labels = self._parse_data(data_or_path, labels)

        # Set up Annoy Engine (only now that n_keys is known)
        self._build_engine(features)

        # Make distribution
        self._make_distribution(features, labels)

        # Filter dataset
        self._set_noise_indices()

        # Set flags
        self.is_fitted = True

        return self.data_without_noise

    def _parse_data(self, data_or_path: Union[pd.DataFrame, str, Path], labels: pd.Series = None):
        """
        Parse data for interval analysis and store it internally

        Parameters
        ----------
        data_or_path : pd.DataFrame or str or Path
            Multi-indexed dataset containing feature (and target) columns
            or path to folder that is correctly structured (see Notes)
        labels : pd.Series, optional
            Multi-indexed dataset containing target columns
        """
        # Verbose
        if self.verbose > 0:
            print('[AutoML] Parsing data for interval analyser.')

        # Parse data
        if isinstance(data_or_path, (str, Path)):
            # Read from path
            features, metadata = merge_logs(data_or_path, self.target)
            labels = features.pop(self.target)
            # Store metadata in self
            self._metadata = metadata
        else:
            assert isinstance(data_or_path, pd.DataFrame), 'Invalid data type'
            # Check if target in `data_or_path`
            if self.target in data_or_path.columns:
                features = data_or_path.drop(self.target, axis=1)
                labels = data_or_path.loc[:, self.target]
            else:
                assert isinstance(labels, pd.Series), 'Invalid data type'
                features = data_or_path

        # Tests
        assert list(features.index.names) == ['log', 'index'], 'Invalid multi-indexed data detected'
        # assert features.index.size == features.index.unique().size, 'Not all indices are unique'
        assert all(features.index == labels.index), 'Indices mismatch: Features and labels cannot be concatenated'
        assert len(features) == len(labels), 'Length mismatch: Features and labels cannot be concatenated'
        assert self.target not in features.columns, 'Target column is already present in feature data'

        # Check data
        assert check_dataframe_quality(features), 'Data quality is insufficient'
        if not check_pearson_correlation(features, labels):
            warnings.warn(UserWarning('Data is correlated. No good results are expected.'))

        # Set name of labels
        if self.target != labels.name:
            warnings.warn('Expected target name does not match the actual name. '
                          f'Name will be fixed from {labels.name} to {self.target}.')
            labels.name = self.target

        # Store data in self
        self._features = features
        self._labels = labels

        # Remove datetime columns
        _date_cols = [col for col in features.columns
                      if pd.api.types.is_datetime64_any_dtype(features[col])]
        if len(_date_cols) != 0:
            features.drop(_date_cols, axis=1, inplace=True)

        # # Normalize
        # self._mins, self._maxs = features.min(), features.max()
        # features = (features - features.min()) / (features.max() - features.min())

        # Change float64 to float32
        #  See: https://github.com/facebookresearch/faiss/issues/461
        for col in features.select_dtypes(include=['float64']).columns:
            features[col] = features[col].astype('float32')

        # Set sizes
        self.n_folders = labels.nunique()
        self.n_files = features.index.get_level_values(0).nunique()
        self.n_samples, self.n_columns = features.shape
        if self.n_neighbors is None:
            self.n_neighbors = min(3 * self.n_samples // self.n_files, 5000)

        return features, labels

    def _build_engine(self, features: pd.DataFrame):
        """
        Builds the ANNOY engine.

        Parameters
        ----------
        features : pd.DataFrame
            Multi-indexed dataset containing feature columns
        """
        # Create engine
        if self.verbose > 0:
            print('[AutoML] Building interval analyser engine.')
        engine = faiss.IndexFlatL2(self.n_columns)

        # Add the data to ANNOY
        engine.add(np.ascontiguousarray(features.values))  # noqa

        # Set class attribute
        self._engine = engine

    def _make_distribution(self, features: pd.DataFrame, labels: pd.Series):
        """
        Given a build K-Nearest Neighbors, returns the label distribution

        Parameters
        ----------
        features : pd.DataFrame
            Multi-indexed dataset containing feature columns
        labels : pd.Series
            Multi-indexed dataset containing target columns
        """
        if self.verbose > 0:
            print('[AutoML] Calculating interval within-class distributions.')

        # Search nearest neighbors for all samples -- has to be iterative for large files -.-
        distribution = []
        for i, row in features.iterrows():
            _, neighbors = self._engine.search(np.ascontiguousarray(row.values.reshape((1, -1))), int(self.n_neighbors))
            match_mask = labels.iloc[neighbors.reshape(-1)] == labels.loc[i]
            distribution.append(pd.Series(match_mask).sum() / self.n_neighbors)

        # Parse into list of lists
        self._distributions = pd.Series(distribution, index=self._labels.index)

    def _set_noise_indices(self):
        """
        This function selects samples given the calculated distributions. It only removes samples from logs which are
        longer (> min_length), and only the samples with lower in-class neighbors.

        One could come up with a fancier logic, using the total dataset samples, the class-balance & sample redundancy.
        """
        # Verbose
        if self.verbose > 0:
            print('[AutoML] Creating filtered dataset')

        # Get in-class means and number of samples for each label
        # label_means = self._get_label_means(labels, self._distributions)
        # label_samples = labels.value_counts()

        # Initialize
        data_index = self._features.index
        noise_indices = []

        # Iterate through labels and see if we should remove values
        for file_id in range(self.n_files):

            # Check length and continue if short
            if sum(data_index.get_level_values(0) == file_id) < self.min_length:
                continue

            # Check distribution and find cut-off
            dist = self._distributions[file_id]
            ind_remove_label = [(file_id, j) for j in np.where(dist < dist.mean())[0]]
            # Extend list to keep track
            noise_indices.extend(ind_remove_label)

            # Verbose
            if len(noise_indices) > 0 and self.verbose > 1:
                filename = self._metadata.iloc[file_id]['file'] \
                    if isinstance(self._metadata, pd.DataFrame) else None
                print(f'[AutoML] Removing {len(ind_remove_label)} samples from `{filename or file_id}`')

        # Set noise indices
        self._noise_indices = noise_indices

    # @staticmethod
    # def _get_label_means(labels, dists):
    #     # Init
    #     label_means = {k: [] for k in labels.unique()}
    #
    #     # Append distributions
    #     for i, d in enumerate(dists):
    #         # Skip failed reads
    #         if i not in labels:
    #             continue
    #
    #         # Extend distribution
    #         label_means[labels[(i, 0)]].extend(d)
    #
    #     # Return means
    #     return {k: np.mean(v) for k, v in label_means.items()}

    # --- Deprecation Properties ---

    @property
    def samples(self):
        warnings.warn(DeprecationWarning('This pseudo-attribute will be removed in a future version. '
                                         'Consider using `n_samples` instead.'))
        return self.n_samples

    @property
    def n_keys(self):
        warnings.warn(DeprecationWarning('This pseudo-attribute will be removed in a future version. '
                                         'Consider using `n_columns` instead.'))
        return self.n_columns



import re
import time
import copy
import random
import itertools
import numpy as np
import pandas as pd
from tqdm import tqdm
from datetime import datetime
from typing import Dict, Union

from ._GridSearch import _GridSearch


# noinspection PyUnresolvedReferences
class BaseGridSearch(_GridSearch):

    def __init__(self, model, *args, **kwargs):
        """
        Basic exhaustive grid search.
        @param model: Model object to optimize
        @param cv: Cross-Validator for scoring
        @param scoring: make_scorer for scoring
        """
        super().__init__(model, *args, **kwargs)
        # Initiate
        self.parsedParams = []
        self.result = []

    def _get_hyper_params(self) -> Dict[str, Union[None, bool, int, float, str, list]]:
        param_values = self._hyper_parameter_values
        param_values.pop('CONDITIONALS', {})  # drop conditionals as they are not supported

        # Drop/adjust some more parameter names which HalvinGridSearch cannot handle
        model_name = type(self.model).__name__
        if model_name in ('BaggingClassifier', 'BaggingRegressor'):
            param_values['max_features'] = ('categorical', [0.5, 0.8, 0.9, 1.0], None)
        elif model_name in ('LGBMRegressor', 'LGBMClassifier'):
            # Someway it cannot handle uniform distributions for this two cases
            param_values['bagging_fraction'] = ('categorical', [0.5, 0.7, 0.9, 1.0], None)
            param_values['feature_fraction'] = ('categorical', [0.5, 0.7, 0.9, 1.0], None)

        # Extract parameter distributions
        params = {}
        for p_name, value in param_values.items():

            # Read out
            p_type = value[0]  # parameter type (str)
            p_args = value[1]  # parameter arguments (list, tuple)
            p_count = value[2] if p_type != 'categorical' else None  # n_samples for np.linspace or np.logspace

            # Sanity checks
            assert len(p_args) == 2 or p_type == 'categorical', \
                'Only categorical parameter can have more/less than two suggest args'

            # Skip parameters when none are wanted
            if p_count == 0:
                continue
            # Suggest parameter given the arguments
            if p_type == 'categorical':
                params[p_name] = p_args
                continue
            elif p_type in ('int', 'uniform'):
                params[p_name] = np.linspace(*p_args, p_count)
            elif p_type in ('logint', 'loguniform'):
                p_args = [np.log(arg) for arg in p_args]  # logarithmise for np.logspace
                params[p_name] = np.logspace(*p_args, p_count)
            else:
                raise NotImplementedError('Invalid parameter specification')
            # Discretize and convert to list
            if 'int' in p_type:
                params[p_name] = params[p_name].astype('int')
            params[p_name] = np.unique(params[p_name]).tolist()

        return params

    def _parse_params(self):
        if len(self.params.items()) > 0:
            k, v = zip(*self.params.items())
            self.parsedParams = [dict(zip(k, v)) for v in itertools.product(*self.params.values())]
            random.shuffle(self.parsedParams)
            print('[GridSearch] Optimizing {}, {}-folds with {} parameter combinations, {} runs.' .format(
                type(self.model).__name__, self.cv.n_splits, len(self.parsedParams),
                len(self.parsedParams) * self.cv.n_splits))
        else:
            self.parsedParams = [{}]

    def fit(self, x, y):
        start_time = time.time()
        # todo check data
        # Parse Params
        model_name = type(self.model).__name__
        is_classification = bool(re.match(r'.*(Classification|Classifier|SVC)', model_name))
        if is_classification:
            if len(np.unique(y)) == 2:
                self.binary = True
            else:
                self.binary = False
        # Convert to Numpy
        if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
            x = np.array(x)
        if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y = np.array(y).reshape((-1))
        self.samples = len(y)

        # Get params
        if self.params is None:
            self.params = self._get_hyper_params()
        self._parse_params()

        # Loop through parameters
        for i, param in tqdm(enumerate(self.parsedParams)):
            # Timeout or trials limit
            if time.time() - start_time > self.timeout or i > self.nTrials:
                return pd.DataFrame(self.result)

            # Scoring and timing vector
            scoring = []
            timing = []

            # Cross Validation
            for train_ind, val_ind in self.cv.split(x, y):
                # Start Timer
                t = time.time()

                # Split data
                x_train, x_val = x[train_ind], x[val_ind]
                y_train, y_val = y[train_ind], y[val_ind]

                # Model training
                model = copy.deepcopy(self.model)
                model.set_params(**param)
                model.fit(x_train, y_train)

                # Results
                scoring.append(self.scoring(model, x_val, y_val))
                timing.append(time.time() - t)

            # Output Printing
            if self.verbose > 0:
                print('[GridSearch][{}] Score: {:.4f} \u00B1 {:.4f} (in {:.1f} seconds) (Best score so-far: {:.4f}'
                      ' \u00B1 {:.4f}) ({} / {})'.format(datetime.now().strftime('%H:%M'), np.mean(scoring),
                                                         np.std(scoring), np.mean(timing), self.best[0], self.best[1],
                                                         i + 1, len(self.parsedParams)))

            self.result.append({
                'date': datetime.today().strftime('%d %b %y'),
                'model': type(model).__name__,
                'mean_objective': np.mean(scoring),
                'std_objective': np.std(scoring),
                'worst_case': np.mean(scoring) - np.std(scoring),
                'params': param,
                'mean_time': np.mean(timing),
                'std_time': np.std(timing),
            })
        return pd.DataFrame(self.result)



import pytest

from Amplo.API.platform import PlatformSynchronizer
from Amplo.Utils.testing import make_production_data

from tests.unit.api import TestAPI


class TestPlatform(TestAPI):

    def test_upload(self):
        """
        This only tests that upload_latest_model doesn't error.
        - Good to check whether the file exists in the cloud.
        - Also needs some cleanup to make sure it doesn't stay there.
        - Also does upload_latest_model truly take the latest?
        - What if latest locally and in the cloud are outdated? It needs to go in the right folder.
        """
        # TODO: implement as soon as platform is out of beta mode
        pytest.skip('Test not yet implemented')

        # Make dummy production data
        issue_dir, kwargs = make_production_data(
            self.sync_dir, team='Demo', machine='Charger 75kW', service='Diagnostics')
        # Upload
        sync = PlatformSynchronizer()
        sync.upload_latest_model(issue_dir, **kwargs)
